#ifndef IST_BBUFFER_H
#define IST_BBUFFER_H

#include <vector>
#include <string>
#include <algorithm>
#include <iostream>
#include <fstream>
#include "vector.h"
#include "ist_conf.h"

namespace ist {


class bstream
{
public:
	virtual ~bstream() {}
	virtual void write(const void *p, size_t s)=0;
	virtual bool read(void *p, size_t s)=0;

	virtual size_t wtell() const=0;
	virtual void wseek(int p)=0;
	virtual void wskip(int p)=0;

	virtual size_t rtell() const=0;
	virtual void rseek(int p)=0;
	virtual void rskip(int p)=0;

	virtual size_t size() const=0;
	virtual bool eof() const=0;
	virtual bool operator!() const=0;
};



class biostream : public bstream
{
public:
	biostream(std::iostream& f) : m_stream(f) {}

	void write(const void *p, size_t s) {
		m_stream.write((char*)p, std::streamsize(s));
	}

	bool read(void *p, size_t s) {
		m_stream.read((char*)p, std::streamsize(s));
		return true;
	}

	size_t wtell() const { return size_t(m_stream.tellp()); }
	void wseek(int p) { m_stream.seekp(p, std::ios::beg); }
	void wskip(int p) { m_stream.seekp(p, std::ios::cur); }

	size_t rtell() const { return size_t(m_stream.tellg()); }
	void rseek(int p) { m_stream.seekg(p, std::ios::beg); }
	void rskip(int p) { m_stream.seekg(p, std::ios::cur); }

	size_t size() const {
		std::streampos cur = m_stream.tellg();
		m_stream.seekg(0, std::ios::end);
		std::streampos filesize = m_stream.tellg();
		m_stream.seekg(cur, std::ios::beg);
		return size_t(filesize);
	}

	bool eof() const { return m_stream.eof(); }
	bool operator!() const { return !m_stream; }



	std::iostream& m_stream;
};



class bbuffer : public bstream, public std::vector<char>
{
public:
	typedef std::vector<char> BaseClass;

	bbuffer() : read_ptr(0), write_ptr(0) {
		reserve(128);
	}

	void write(const void *p, size_t s) {
		if(write_ptr+s>size())
			resize(write_ptr+s);
		memcpy(ptr()+write_ptr, p, s);
		write_ptr+=s;
	/*	insert(end(), (const char*)p, (const char*)p+s);
		write_ptr+=s;
	*/
	}

	bool read(void *p, size_t s) {
		if(read_ptr+s>size()) {
			read_ptr = size();
			return false;
		}
		memcpy(p, ptr()+read_ptr, s);
		read_ptr+=s;
		return true;
	}

	size_t wtell() const { return write_ptr; }
	void wseek(int i) { write_ptr=size_t(i); }
	void wskip(int i) { write_ptr+=i; }

	size_t rtell() const { return read_ptr; }
	void rseek(int i) { read_ptr=size_t(i); }
	void rskip(int i) { read_ptr+=i; }

	size_t size() const { return BaseClass::size(); }
	bool eof() const { return read_ptr>=size(); }
	bool operator!() const { return read_ptr>=size(); }


	void clear() {
		std::vector<char>::clear();
		read_ptr = 0;
		write_ptr = 0;
	}

	bool load(const char *path) {
		std::ifstream in(path, std::ios::binary);
		if(!in)
			return false;
		size_t size = 0;
		in.seekg(0, std::ios::end);
		size = in.tellg();
		resize(size);
		in.seekg(0, std::ios::beg);
		in.read(ptr(), std::streamsize(size));
		return true;
	}

	bool save(const char *path) const {
		std::ofstream out(path, std::ios::binary);
		if(!out)
			return false;
		out.write(ptr(), std::streamsize(size()));
		return true;
	}


	char* ptr() { return &(*this)[0]; }
	const char* ptr() const { return &(*this)[0]; }
	std::string str() const { return std::string(ptr(), size()); }

private:
	size_t read_ptr;
	size_t write_ptr;
};






class barchive
{
public:
	virtual ~barchive() {}
	virtual void io(void *p, size_t s)=0;

	virtual bool isReader() { return false; }
	virtual bool isWriter() { return false; }
};

class bwriter : public barchive
{
public:
	bwriter(bstream& s) : stream(s) {}
	void io(void *p, size_t s) { stream.write(p, s); }
	virtual bool isWriter() { return true; }

	bstream& stream;
};

class breader : public barchive
{
public:
	breader(bstream& s) : stream(s) {}
	void io(void *p, size_t s) { stream.read(p, s); }
	virtual bool isReader() { return true; }

	bstream& stream;
};


} // ist



template<class T>
inline T& _CE(T& v) {
#ifdef _BIG_ENDIAN
	std::reverse((char*)(&v), (char*)(&v)+sizeof(T));
#endif
	return v;
}

/*
	bstream
*/

inline void operator+=(ist::bbuffer& b, const ist::bbuffer& a) { b.write(a.ptr(), a.size()); }

inline ist::bstream& operator<<(ist::bstream& b, char v)          { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, unsigned char v) { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, short v)         { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, unsigned short v){ b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, int v)           { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, unsigned int v)  { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, float v)         { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, double v)        { b.write(&v, sizeof(v)); return b; }
inline ist::bstream& operator<<(ist::bstream& b, bool v)          { b.write(&v, sizeof(v)); return b; }

inline ist::bstream& operator<<(ist::bstream& b, const std::string& v) {
	b << v.size();
	b.write(v.c_str(), v.size());
	return b;
}
inline ist::bstream& operator<<(ist::bstream& b, const ist::vector4& v) {
	b << v.x << v.y << v.z;
	return b;
}
inline ist::bstream& operator<<(ist::bstream& b, const ist::bbuffer& v) {
	b << v.size();
	b.write(v.ptr(), v.size());
	return b;
}


inline ist::bstream& operator>>(ist::bstream& b, char& v)          { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, unsigned char& v) { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, short& v)         { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, unsigned short& v){ b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, int& v)           { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, unsigned int& v)  { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, float& v)         { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, double& v)        { b.read(&v, sizeof(v)); return b; }
inline ist::bstream& operator>>(ist::bstream& b, bool& v)          { b.read(&v, sizeof(v)); return b; }

inline ist::bstream& operator>>(ist::bstream& b, std::string& v) { 
	size_t size = 0;
	b >> size;
	v.resize(size);
	for(size_t i=0; i<size; ++i)
		b.read(&v[i], 1);
	return b;
}
inline ist::bstream& operator>>(ist::bstream& b, ist::vector4& v) {
	b >> v.x >> v.y >> v.z;
	return b;
}
inline ist::bstream& operator>>(ist::bstream& b, ist::bbuffer& v) {
	size_t size = 0;
	b >> size;
	v.clear();

	v.resize(size);
	b.read(v.ptr(), size);
/*	char c;
	for(size_t i=0; i<size && !!b; ++i) {
		b >> c;
		v << c;
	}
*/
	return b;
}



inline std::iostream& operator<<(std::iostream& ios, ist::bbuffer& v) {
	size_t size = v.size();
	ios.write((char*)&size, sizeof(size_t));
	ios.write(v.ptr(), std::streamsize(v.size()));
	return ios;
}

inline std::iostream& operator>>(std::iostream& ios, ist::bbuffer& v) {
	v.clear();
	size_t size = 0;
	ios.read((char*)&size, std::streamsize(sizeof(size_t)));
	v.clear();
	v.resize(size);
	ios.read(v.ptr(), size);
	return ios;
}


/*
	barchive
*/

inline ist::barchive& operator & (ist::barchive& b, char& v)  { b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, short& v) { b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, int& v)   { b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, size_t& v){ b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, float& v) { b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, double& v){ b.io(&v, sizeof(v)); return b; }
inline ist::barchive& operator & (ist::barchive& b, bool& v)  { b.io(&v, sizeof(v)); return b; }

inline ist::barchive& operator & (ist::barchive& b, std::string& v) {
	size_t size = v.size();
	b & size;
	if(b.isWriter()) {
		b.io(const_cast<char*>(v.c_str()), v.size());
	}
	else if(b.isReader()) {
		v.resize(size);
		for(size_t i=0; i<size; ++i)
			b.io(&v[i], 1);
	}
	return b;
}
inline ist::barchive& operator & (ist::barchive& b, ist::vector4& v) {
	b & v.x & v.y & v.z;
	return b;
}
inline ist::barchive& operator & (ist::barchive& b, ist::bbuffer& v) {
	size_t size = v.size();
	b & size;
	if(b.isReader())
		v.resize(size);
	b.io(v.ptr(), v.size());
	return b;
}


#endif
