#ifndef IST_GRID_H
#define IST_GRID_H

#include <vector>
#include "ist/vector.h"
#include "iterator.h"
#include "array2.h"

namespace ist {


template<
	class T,
	class Cont = std::vector<T>
>
class grid_structure2
{
public:
	typedef T value_type;



	class block_elem : public Cont
	{
	public:
		typedef typename Cont::value_type value_type;
		typedef typename Cont::iterator iterator;
		float radius;
		ist::vector4 center;
		
		block_elem() : radius(0) {}
		const ist::vector4& getPos() const { return center; }
		float getRadius() const { return radius; }
	};

	typedef array2<block_elem> block_cont;
	typedef MultiContainerIterator<block_elem> block_iter;
	typedef std::map<value_type*, block_elem*> mark_cont;


	class point2
	{
	public:
		point2(int x_=0, int z_=0) : x(x_), z(z_) {}
		int x, z;
	};




	grid_structure2() : x_div(0), z_div(0) {
		cell = ist::vector4(100.0f, 100.0f, 100.0f); // 0Zh~ 
	}

	grid_structure2(const ist::vector2& mn, const ist::vector2& mx, int xd, int zd) {
		resize(mn, mx, xd, zd);
	}

	bool empty() { return blocks.empty(); }

	void resize(const ist::vector2& mn, const ist::vector2& mx, int xd, int zd) {
		resize(ist::vector4(mn.x, 0, mn.y), ist::vector4(mx.x, 0, mx.y), xd, zd);
	}

	void resize(const ist::vector4& mn, const ist::vector4& mx, int xd, int zd) {
		min = mn;
		max = mx;
		x_div = xd;
		z_div = zd;
		area = max-min;
		cell.x = area.x/x_div;
		cell.z = area.z/z_div;

		blocks.resize(x_div, z_div);
		for(int i=0; i<z_div; ++i) {
			for(int j=0; j<x_div; ++j) {
				ist::vector4 b_min(min.x+cell.x*j, 0, min.z+cell.z*i);
				ist::vector4 b_max(min.x+cell.x*(j+1), 0, min.z+cell.z*(i+1));
				blocks.at(j, i).center = (b_min+b_max)/2;
				blocks.at(j, i).radius = (b_max-b_min).norm()/2;
			}
		}
	}

	bool insert(const ist::vector4& pos, const value_type& value) {
		block_elem& block = getBlockByPosition(pos);
		if(&block==&null_block)
			return false;

		block.push_back(value);
		return true;
	}

	bool move(const ist::vector4& before, const ist::vector4& after, const value_type& value) {
		if(before==after)
			return false;

		block_elem& bc = getBlockByPosition(before);
		block_elem& ac = getBlockByPosition(after);
		if(&bc==&ac)
			return false;

		if(&bc!=&null_block)
			bc.erase(std::find(bc.begin(), bc.end(), value));
		if(&ac==&null_block)
			return false;

		ac.push_back(value);
		return true;
	}

	bool erase(const ist::vector4& pos, const value_type& value) {
		block_elem& block = getBlockByPosition(pos);
		if(&block==&null_block)
			return false;

		typename block_elem::iterator p = std::find(block.begin(), block.end(), value);
		if(p!=block.end()) {
			block.erase(p);
			return true;
		}
		return false;
	}

	void clear() {
		for(typename block_cont::iterator p=blocks.begin(); p!=blocks.end(); ++p)
			p->clear();
	}


	int getXDivision() const { return x_div; }
	int getZDivision() const { return z_div; }
	const ist::vector4& getMinPos() const { return min; }
	const ist::vector4& getMaxPos() const { return max; }

	point2 getCoordByPosition(const ist::vector4& pos) const {
		ist::vector4 rel = pos-min;
		return point2(int(rel.x/cell.x), int(rel.z/cell.z));
	}

	block_elem& getBlockByCoord(const point2& coord) {
		int x=coord.x, z=coord.z;
		if(x<0 || z<0 || x>=x_div || z>=z_div)
			return null_block;
		return blocks.at(x, z);
	}

	block_elem& getBlockByPosition(const ist::vector4& pos) {
		return getBlockByCoord( getCoordByPosition(pos) );
	}

	block_iter& getBlockByCoordBox(const point2& min_c, const point2& max_c, block_iter& iter=_default_iter()) {
		iter.clear();
		for(int j=min_c.z; j<=max_c.z; ++j) {
			for(int k=min_c.x; k<=max_c.x; ++k) {
				block_elem& be = getBlockByCoord(point2(k, j));
				if(&be!=&null_block)
					iter.add(be);
			}
		}
		iter.reset();
		return iter;
	}

	block_iter& getBlocksByBox(const ist::vector4& r_min, const ist::vector4& r_max, block_iter& iter=_default_iter()) {
		return getBlockByCoordBox(getCoordByPosition(r_min), getCoordByPosition(r_max), iter);
	}

	block_iter& getBlocksByBorder(const ist::vector4& r_min, int level, block_iter& iter=_default_iter()) {
		point2 center = getCoordByPosition(r_min);
		point2 c_min(center.x-level, center.z-level);
		point2 c_max(center.x+level, center.z+level);
		return getBlockByCoordBox(c_min, c_max, iter);
	}

	block_iter& getBlocksByRadius(const ist::vector4& pos, float radius, block_iter& iter=_default_iter()) {
		ist::vector4 r_min(pos.x-radius, 0, pos.z-radius);
		ist::vector4 r_max(pos.x+radius, 0, pos.z+radius);
		return getBlocksByBox(r_min, r_max, iter);
	}

	block_iter& getAllBlocks(block_iter& iter=_default_iter()) {
		iter.clear();
		for(block_cont::iterator p=blocks.begin(); p!=blocks.end(); ++p)
			iter.add(*p);
		return iter;
	}


private:
	static block_iter& _default_iter() {
		static block_iter i;
		return i;
	}

	block_cont blocks;
	block_elem null_block;
	mark_cont marks;
	int x_div;
	int z_div;
	ist::vector4 min;
	ist::vector4 max;
	ist::vector4 area;
	ist::vector4 cell;
};


} // namespace ist 

#endif // IST_GRID_H
