#ifndef IST_I3D_H
#define IST_I3D_H

#include <iostream>
#include <fstream>
#include <vector>
#include <list>
#include <map>
#include "ist_conf.h"
#include "vector.h"
#include "bitmap.h"
#include "bstream.h"
#include "gl_include.h"

#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif


namespace ist {

/*
	Diagram
*/

struct Sphere
{
	Sphere() : radius(0) {}
	Sphere(const vector4& p, float r) : pos(p), radius(r) {}

	ist::vector4 pos;
	float radius;
};


struct AABox
{
	AABox() {}
	AABox(const ist::vector4 &min_, const vector4 &max_) : min(min_), max(max_) {}

	AABox getAABB(const ist::matrix44 &trans) const;
	bool overlap(const AABox& bb) const;
	bool innerPoint(const Sphere&) const;
	bool innerPoint(const ist::vector4& pos, float radius=0) const;
	int innerTriangle(const ist::vector4& v1, const ist::vector4& v2, const ist::vector4& v3) const;

	ist::vector4 min, max;
};




/*
	CollisionDetector
*/

enum {
	CM_FALSE,
	CM_TRUE,
	CM_POINT,
	CM_LINE,
	CM_POLYGON,
	CM_POLYGON2D,
};

class DSO_CLASS ICollisionModel : public ist::Object
{
public:
	virtual ~ICollisionModel()=0;
	virtual int getType() const=0;
	virtual void transform(const ist::matrix44&) {}
};

class DSO_CLASS IPointCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_POINT; }
	virtual float getRadius() const=0;
	virtual const ist::vector4& getPos() const=0;
	virtual const AABox& getBoundingBox() const=0;
};

class DSO_CLASS ILineCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_LINE; }
	virtual float getRadius() const=0;
	virtual const ist::vector4& getTail() const=0;
	virtual const ist::vector4& getHead() const=0;
	virtual const AABox& getBoundingBox() const=0;
};

class DSO_CLASS IPolygonCollisionModel : public ICollisionModel
{
public:
	typedef std::vector<ist::vector4> vertex_cont;
	typedef std::vector<ist::vector4> normal_cont;
	typedef std::vector<int>     index_cont;

	int getType() const { return CM_POLYGON; }
	virtual const vertex_cont& getVertex() const=0;
	virtual const normal_cont& getNormal() const=0;
	virtual const index_cont&  getIndex() const=0;
	virtual const AABox& getBoundingBox() const=0;
};

class DSO_CLASS IPolygon2DCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_POLYGON2D; }
	virtual size_t getNumVertex() const=0;
	virtual const ist::vector2& getVertex(int) const=0;
	virtual float getTop() const=0;
	virtual float getBottom() const=0;
};




class DSO_CLASS NullCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_FALSE; }
};

class DSO_CLASS AlmightyCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_TRUE; }
};


class DSO_CLASS SimplePointCollisionModel : public IPointCollisionModel
{
public:
	SimplePointCollisionModel() : radius(0) {}
	SimplePointCollisionModel(const ist::vector4& p, float r) : pos(p), radius(r) { setBoundingBox(); }
	float getRadius() const { return radius; }
	const ist::vector4& getPos() const { return pos; }
	const AABox& getBoundingBox() const { return bb; }

	void setPos(const vector4& v) { pos=v; setBoundingBox(); }
	void setRadius(float r) { radius=r; setBoundingBox(); }
	void setBoundingBox() {
		bb.min.x = pos.x-radius;
		bb.min.y = pos.y-radius;
		bb.min.z = pos.z-radius;
		bb.max.x = pos.x+radius;
		bb.max.y = pos.y+radius;
		bb.max.z = pos.z+radius;
	}

protected:
	ist::vector4 pos;
	float radius;
	AABox bb;
};

class DSO_CLASS SimpleLineCollisionModel : public ILineCollisionModel
{
public:
	SimpleLineCollisionModel() : radius(0) {}
	float getRadius() const { return radius; }
	const ist::vector4& getHead() const { return head; }
	const ist::vector4& getTail() const { return tail; }
	const AABox& getBoundingBox() const { return bb; }

	void setHead(const vector4& v) { head=v; setBoundingBox(); }
	void setTail(const vector4& v) { tail=v; setBoundingBox(); }
	void setRadius(float r) { radius=r; setBoundingBox(); }
	void setBoundingBox() {
		bb.min.x = std::min<float>(head.x+radius, tail.x+radius);
		bb.min.y = std::min<float>(head.y+radius, tail.y+radius);
		bb.min.z = std::min<float>(head.z+radius, tail.z+radius);
		bb.max.x = std::max<float>(head.x+radius, tail.x+radius);
		bb.max.y = std::max<float>(head.y+radius, tail.y+radius);
		bb.max.z = std::max<float>(head.z+radius, tail.z+radius);
	}

protected:
	ist::vector4 head, tail;
	float radius;
	AABox bb;
};

class DSO_CLASS SimplePolygonCollisionModel : public IPolygonCollisionModel
{
public:
	SimplePolygonCollisionModel() {}
	const vertex_cont& getVertex() const { return vertex; }
	const normal_cont& getNormal() const { return normal; }
	const index_cont&  getIndex() const  { return index; }
	const AABox& getBoundingBox() const  { return bb; }

	void setBoundingBox() {
		bb.min = vertex.front();
		bb.max = vertex.front();
		for(vertex_cont::iterator p=vertex.begin(); p!=vertex.end(); ++p) {
			bb.min.x = std::min<float>(bb.min.x, p->x);
			bb.min.y = std::min<float>(bb.min.y, p->y);
			bb.min.z = std::min<float>(bb.min.z, p->z);
			bb.max.x = std::max<float>(bb.max.x, p->x);
			bb.max.y = std::max<float>(bb.max.y, p->y);
			bb.max.z = std::max<float>(bb.max.z, p->z);
		}
	}

	vertex_cont vertex;
	normal_cont normal;
	index_cont  index;
	AABox bb;
};

class DSO_CLASS SimplePolygon2DCollisionModel : public IPolygon2DCollisionModel
{
public:
	SimplePolygon2DCollisionModel() : top(1), bottom(-1) {}
	size_t getNumVertex() const { return vertex.size(); }
	const ist::vector2& getVertex(int i) const { return vertex[i]; }
	float getTop() const    { return top; }
	float getBottom() const { return bottom; }

	typedef std::vector<ist::vector2> vertex_cont;
	vertex_cont vertex;
	float top;
	float bottom;
};



class DSO_CLASS FrustumCollisionModel : public SimplePolygonCollisionModel
{
public:
	FrustumCollisionModel() :
		z_near(0),
		z_far(1500.0f)
	{
		vertex.resize(5);
		normal.resize(6);
		index.resize(18);

		index[0]=0; index[1]=1; index[2]=2;
		index[3]=0; index[4]=2; index[5]=3;
		index[6]=0; index[7]=3; index[8]=4;
		index[9]=0; index[10]=4; index[11]=1;
		index[12]=1; index[13]=4; index[14]=3;
		index[15]=3; index[16]=2; index[17]=1;
	}

	void setup(const vector4& _0, const vector4& _1, const vector4& _2, const vector4& _3, const vector4& _4) {

		vertex[0] = _0;
		vertex[1] = _1;
		vertex[2] = _2;
		vertex[3] = _3;
		vertex[4] = _4;
		for(size_t i=0; i<index.size()/3; ++i) {
			normal[i] = vector4().cross(
				vertex[index[i*3+1]]-vertex[index[i*3]], vertex[index[i*3+2]]-vertex[index[i*3]]).normalize();
		}

		setBoundingBox();
	}

	bool test(const vector4& center, float r) const {
		vector4 rel_pos = center-vertex[0];
		float dist = normal[4].dot(rel_pos);

		if(vertex[1]==vertex[2] || vertex[2]==vertex[3] || dist<z_near-r || dist>z_far+r)
			return false;
		if(normal[0].dot(rel_pos)>r || normal[2].dot(rel_pos)>r || normal[3].dot(rel_pos)>r || normal[1].dot(rel_pos)>r)
			return false;
		return true;
	}

	bool test(const ICollisionModel& cm) const {
		if(cm.getType()==CM_TRUE) {
			return true;
		}
		else if(cm.getType()==CM_POINT) {
			const IPointCollisionModel& pcm = static_cast<const IPointCollisionModel&>(cm);
			return test(pcm.getPos(), pcm.getRadius());
		}
		else if(cm.getType()==CM_LINE) {
			const ILineCollisionModel& lcm = static_cast<const ILineCollisionModel&>(cm);
			return test(lcm.getHead(), lcm.getRadius());
		}

		return false;
	}

	float z_near;
	float z_far;
};




class DSO_CLASS CollisionDetector : public ist::Object
{
public:
	CollisionDetector();
	bool detect(const ICollisionModel& base, const ICollisionModel& target);

	bool result;
	float distance;
	ist::vector4 collide_pos;
	ist::vector4 collide_normal;

private:
	template <class T, class U>
	bool detect_(const T &base, const U &target);
};



/*
	GLInfo
*/

class DSO_CLASS GLInfo
{
public:
	void print(std::iostream& out);
};



/*
	Matrix
*/

DSO_EXPORT void ScreenMatrix(); // WnEBhEɈv 
DSO_EXPORT void ScreenMatrixIV(); // ㉺t 


class DSO_CLASS MatrixSaver : public ist::Object
{
public:

// GL_MODELVIEW_MATRIX | GL_PROJECTION_MATRIX | GL_TEXTURE_MATRIX
	MatrixSaver(int matrix_type = GL_MODELVIEW_MATRIX);
	~MatrixSaver();
	const matrix44& getMatrix() const;

private:
	int matrix_mode;
	ist::matrix44 matrix;
};



/*
	Material
*/

class DSO_CLASS Material : public ist::Object
{
public:
	Material();
	virtual void assign() const;

public:
	ist::vector4 ambient;
	ist::vector4 diffuse;
	ist::vector4 specular;
	ist::vector4 emission;
	float shininess;
};


class DSO_CLASS Color : public ist::Object
{
public:
	Color();
	Color(float r, float g, float b, float a);
	virtual void assign() const;

public:
	ist::vector4 color;
};


/*
	Light
*/

class DSO_CLASS Light : public ist::Object
{
private:
	Light(const Light&);
	Light& operator=(const Light&);

public:
	Light();
	~Light();

	void enable() const;
	void disable() const;
	bool isValid() const;

public:
	ist::vector4 position;
	ist::vector4 ambient;
	ist::vector4 diffuse;
	ist::vector4 specular;
};


/*
	Fog
*/

class DSO_CLASS Fog : public ist::Object
{
public:
	Fog();
	~Fog();

	void enable() const;
	void disable() const;

public:
	ist::vector4 position;
	ist::vector4 color;
	float start;
	float end;
};



/*
	Camera
*/


class DSO_CLASS ViewFrustum : public SimplePolygonCollisionModel
{
public:
	ViewFrustum();
	void setup(const vector4& pos_, const vector4& target4, const vector3& up, float z_near_, float z_far_, float fovy, float aspect);
	int test(const vector4& center, float r) const;
	int test(const ICollisionModel& cm) const;

	ist::vector4 pos;
	ist::vector3 front;
	ist::vector3 top;
	ist::vector3 bottom;
	ist::vector3 right;
	ist::vector3 left;
	float z_near;
	float z_far;
};



class DSO_CLASS Camera : public ist::Object
{
public:
	Camera();
	virtual ~Camera();
	virtual void look()=0;

	ist::vector4 position;
	ist::vector4 target;
	ist::vector3 direction;
	float z_near;
	float z_far;
};


class DSO_CLASS OrthographicCamera : public Camera
{
public:
	OrthographicCamera();
	virtual void look();
	virtual void setScreen(float l, float r, float b, float t);

	float left;
	float right;
	float bottom;
	float top;
};

class DSO_CLASS PerspectiveCamera : public Camera
{
public:
	PerspectiveCamera();
	virtual void look();
	virtual void setViewFrustum();
	virtual const ViewFrustum& getViewFrustum() const;

	ViewFrustum view_frustum;
	float fovy;
	float aspect;
};







/*
	Texture
*/

inline size_t power_of_two(size_t size) {
	size_t i=1;
	while(i<size)
		i*=2;
	return i;
}

struct TextureParam
{
	int warp_s;
	int warp_t;
	int mag_filter;
	int min_filter;
	int env_mode;

	TextureParam() {
	}
};

class DSO_CLASS ITexture : public ist::Object
{
public:
	virtual ~ITexture() {}

	virtual bool load(const char *filename, size_t cell_width=0, size_t cell_height=0) =0;
	virtual bool load(const Bitmap& src, size_t cell_width=0, size_t cell_height=0) =0;

	virtual void clear()=0;
	virtual void assign(int index=0) const =0;
	virtual bool operator !()  const=0;

	virtual unsigned int getNumCell() const=0;
};

// Rs[s 
class DSO_CLASS Texture : public ITexture
{
private:
	Texture(const Texture&);
	Texture& operator =(const Texture&);

	void setDefaultParam();
	void setParam(unsigned int) const;
	void generate(unsigned int&, const Bitmap&);

	int _dimension;
	int _type;
	int _warp_s;
	int _warp_t;
	int _mag_filter;
	int _min_filter;
	int _env_mode;

	std::vector<unsigned int> _tex_name;

public:
	Texture();
	Texture(const char *filename, size_t cell_width=0, size_t cell_height=0);
	Texture(const Bitmap& src, size_t cell_width=0, size_t cell_height=0);
	virtual ~Texture();

	virtual bool load(const char *filename, size_t cell_width=0, size_t cell_height=0);
	virtual bool load(const Bitmap& src, size_t cell_width=0, size_t cell_height=0);
	virtual void clear();
	virtual void assign(int index=0) const;
	virtual bool operator !()  const { return _tex_name.empty(); }

	virtual unsigned int getName(int index=0) { return _tex_name[index]; }
	virtual size_t getNumCell() const{ return _tex_name.size(); }
	void setDimension(int a) { _dimension=a; }
	void setType(int a)      { _type=a; }
	void setMagFilter(int a) { _mag_filter=a; }
	void setMinFilter(int a) { _min_filter=a; }
	void setWarpS(int a)     { _warp_s=a; }
	void setWarpT(int a)     { _warp_t=a; }
	void setEnvMode(int a)   { _env_mode=a; }

};


/*
	PolygonModel
*/

struct vertex_t
{
	ist::vector4 position;
	size_t bone_id[8];
	float weight[8];

	vertex_t() {
		for(size_t i=0; i<8; ++i) {
			bone_id[i] = 0;
			if(i==0)
				weight[i] = 100.0f;
			else
				weight[i] = 0.0f;
		}
	}
};

class DSO_CLASS IPolygonChunc : public ist::Object
{
public:
	typedef std::vector<vertex_t> vertex_cont;
	typedef std::vector<ist::vector3>  position_cont;
	typedef std::vector<ist::vector3>  normal_cont;
	typedef std::vector<ist::vector2>  texcoord_cont;
	typedef std::vector<ist::matrix44> bone_cont;
	typedef std::vector<size_t> index_cont;

	virtual ~IPolygonChunc() {}
	virtual void draw() const=0;
	virtual const vertex_cont& getVertex() const=0;
	virtual const normal_cont& getNormal() const=0;
	virtual const index_cont&  getIndex() const=0;
	virtual const Sphere& getBSphere() const=0;
	virtual const AABox&  getBBox() const=0;
};

class DSO_CLASS IPolygonModel : public ist::Object
{
public:
	typedef boost::intrusive_ptr<IPolygonChunc> chunc_ptr;

	virtual ~IPolygonModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw() const=0;

	virtual size_t getNumChunc()=0;
	virtual chunc_ptr getChunc(int i)=0;
	virtual chunc_ptr getChunc(const std::string& chuncname)=0;
};


class DSO_CLASS BPolygonChunc : public IPolygonChunc
{
public:
	BPolygonChunc();
	virtual ~BPolygonChunc();

	virtual void clear();
	virtual void draw() const;

	virtual void setup();
	virtual void setNormal();
	virtual void setBoundingVolume();

	virtual const vertex_cont& getVertex() const { return vertex; }
	virtual const normal_cont& getNormal() const { return normal; }
	virtual const index_cont&  getIndex() const { return index; }
	virtual const Sphere& getBSphere() const { return bsphere; }
	virtual const AABox&  getBBox() const { return bbox; }


	index_cont  index;
	vertex_cont vertex;
	bone_cont   bone;

	position_cont vvertex;
	normal_cont   normal; // ʖ̖@ 
	normal_cont   vnormal; // _̖@ 
	texcoord_cont vtexcoord;
	AABox         bbox;
	Sphere        bsphere;
	size_t face_num;
	float facet;
};


class DSO_CLASS BPolygonModel : public IPolygonModel
{
public:
	typedef boost::intrusive_ptr<BPolygonChunc> cchunc_ptr;

	BPolygonModel() {}
	BPolygonModel(const std::string& filename) { load(filename); }

	virtual ~BPolygonModel() {}
	virtual bool load(const std::string& filename);
	virtual void draw() const;
	virtual size_t getNumChunc() { return chuncs.size(); }
	virtual chunc_ptr getChunc(int i);
	virtual chunc_ptr getChunc(const std::string& chuncname);

	virtual void addChunc(cchunc_ptr p, const std::string& l) { chuncs.push_back(p); labels.push_back(l); }
	virtual void setup();

	virtual BPolygonChunc* createChunc() { return new BPolygonChunc(); }

private:
	typedef std::list<cchunc_ptr> chunc_cont;
	typedef std::list<std::string> label_cont;
	chunc_cont chuncs;
	label_cont labels;

protected:
	bool loadMQO(const std::string& filename);
	bool loadOBJ(const std::string& filename);
};



// TBasePolygonChuncpclass
template<class T>
class TPolygonModel : public BPolygonModel
{
public:
	typedef T chunc_type;

	TPolygonModel() {}
	TPolygonModel(const std::string& filename) { load(filename); }

	virtual chunc_type* createChunc() { return new chunc_type(); }
};




// OpenGLŕ`ff[^ 
class DSO_CLASS PolygonChunc : public BPolygonChunc
{
public:
	typedef std::vector<float> vertex_array_cont;

	PolygonChunc();
	virtual ~PolygonChunc();

	virtual void draw() const;
	virtual void setup();

	vertex_array_cont vertex_array;
};


// VBOgff[^ 
// Rs[s 
class DSO_CLASS VBOPolygonChunc : public PolygonChunc
{
public:
	VBOPolygonChunc();
	virtual ~VBOPolygonChunc();

	virtual void draw() const;
	virtual void setup();

private:
	unsigned int vbo_id;

	VBOPolygonChunc(const VBOPolygonChunc&);
	const VBOPolygonChunc& operator=(const VBOPolygonChunc&);
};

typedef TPolygonModel<PolygonChunc> PolygonModel;
typedef TPolygonModel<VBOPolygonChunc> VBOPolygonModel;





/*
	MorphModel
*/

class DSO_CLASS IMorphModel : public ist::Object
{
public:
	virtual ~IMorphModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw(const std::string& clipname, float t) const=0;

	virtual bool isExistClip(const std::string& clipname) const=0;
	virtual float beginFrame(const std::string& clipname) const=0;
	virtual float endFrame(const std::string& clipname) const=0;
	virtual const Sphere& getBSphere() const=0;
	virtual const AABox&  getBBox() const=0;
};

class DSO_CLASS BMorphModel : public IMorphModel
{
public:
	typedef std::vector<ist::vector3> vertex_cont;
	typedef std::vector<ist::vector2> texcoord_cont;
	typedef std::vector<size_t>       index_cont;

	class vertex_sequence : public std::map<float, vertex_cont>
	{
	public:
		typedef std::pair<const_iterator, const_iterator> range_t;

		float beginFrame() const {
			if(empty()) return 0;
			return begin()->first;
		}

		float endFrame() const {
			if(empty()) return 0;
			return (--end())->first;
		}

		bool changeFrame(float before, float after) {
			iterator p = find(before);
			if(p==end())
				return false;

			(*this)[after] = p->second;
			erase(p);
			return true;
		}

		range_t range(float f) const {
			return equal_range(f);
		}
		void morph(float t, vertex_cont& t_vertex) const;
	};

	typedef std::map<std::string, vertex_sequence> vertex_clip;
	typedef vertex_clip::const_iterator iterator;


	BMorphModel();
	BMorphModel(const std::string& filename) { load(filename); }
	virtual ~BMorphModel();
	virtual void draw(const std::string& clipname, float t) const;
	virtual void draw(iterator p, float t) const;
	virtual void setup();
	virtual bool load(const std::string& filename);
	virtual bool save(const std::string& filename);
	virtual void clear();

	iterator find(const std::string& clipname) { return vtx_clip.find(clipname); }
	iterator begin() { return vtx_clip.begin(); }
	iterator end()   { return vtx_clip.end(); }

	float beginFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.beginFrame();
	}

	float endFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.endFrame();
	}

	bool isExistClip(const std::string& clipname) const { return vtx_clip.find(clipname)!=vtx_clip.end(); }
	const Sphere& getBSphere() const{ return bsphere; }
	const AABox&  getBBox() const   { return bbox; }


	AABox bbox;
	Sphere bsphere;
	vertex_clip   vtx_clip;
	texcoord_cont tex_array;
	index_cont    idx_array;
	unsigned int  num_vertex;
};



class DSO_CLASS ICachedMorphModel : public ist::Object
{
public:
	virtual ~ICachedMorphModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw(const std::string& clipname, int t) const=0;

	virtual bool isExistClip(const std::string& clipname) const=0;
	virtual int beginFrame(const std::string& clipname) const=0;
	virtual int endFrame(const std::string& clipname) const=0;
	virtual const Sphere& getBSphere() const=0;
	virtual const AABox&    getBBox() const=0;
};

class DSO_CLASS BCachedMorphModel : public ICachedMorphModel
{
public:
	typedef std::vector<vector3> vertex_cont;
	typedef std::vector<vector3> normal_cont;
	typedef std::vector<vector2> texcoord_cont;
	typedef std::vector<size_t>  index_cont;

	class vertex_sequence
	{
	public:
		vertex_sequence() : begin_frame(0), end_frame(0) {}
		bool empty() const     { return vertex.empty() || normal.empty(); }
		int beginFrame() const { return begin_frame; }
		int endFrame() const   { return end_frame; }

		int begin_frame;
		int end_frame;
		std::vector<vertex_cont> vertex;
		std::vector<normal_cont> normal;
	};

	typedef std::map<std::string, vertex_sequence> vertex_clip;
	typedef vertex_clip::const_iterator iterator;


	BCachedMorphModel();
	BCachedMorphModel(const std::string& filename, bool en=true);
	virtual ~BCachedMorphModel();
	virtual void draw(const std::string& clipname, int t) const;
	virtual void draw(iterator p, int t) const;

	virtual void setup();
	virtual bool load(const std::string& filename);
	virtual void setVertex(vertex_cont& vtx_array, vertex_cont& vvtx_array);
	virtual void setNormal(vertex_cont& vtx_array, normal_cont& vnml_array);
	virtual void clear();
	virtual void enableNormal(bool en) { enable_normal = en; }

	iterator find(const std::string& clipname) { return vtx_clip.find(clipname); }
	iterator begin() { return vtx_clip.begin(); }
	iterator end()   { return vtx_clip.end(); }

	int beginFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.beginFrame();
	}

	int endFrame(const std::string& clipname) const {
		vertex_clip::const_iterator p = vtx_clip.find(clipname);
		if(p==vtx_clip.end())
			return 0;
		return p->second.endFrame();
	}

	bool isExistClip(const std::string& clipname) const { return vtx_clip.find(clipname)!=vtx_clip.end(); }
	const Sphere& getBSphere() const { return bsphere; }
	const AABox&  getBBox() const { return bbox; }


	AABox bbox;
	Sphere bsphere;
	vertex_clip   vtx_clip;
	texcoord_cont tex_array;
	index_cont    idx_array;
	unsigned int  num_vertex;
	float facet;
	bool enable_normal;
};




class DSO_CLASS MorphModel : public BMorphModel
{
public:
	MorphModel() {}
	MorphModel(const std::string& filename) { load(filename); }
	virtual void draw(const std::string& clipname, float t) const { BMorphModel::draw(clipname, t); }
	virtual void draw(iterator p, float t) const;
};

/*
class DSO_CLASS VSHMorphModel : public BMorphModel
{
public:
	VSHMorphModel() {}
	VSHMorphModel(const std::string& filename) { load(filename); }
	void draw(iterator p, float t) const;
};
*/


class DSO_CLASS CachedMorphModel : public BCachedMorphModel
{
public:
	CachedMorphModel() {}
	CachedMorphModel(const std::string& filename, bool en=true) { enableNormal(en); load(filename); }
	virtual void draw(const std::string& clipname, int t) const { BCachedMorphModel::draw(clipname, t); }
	virtual void draw(iterator p, int t) const;
};



/*
	Kanji
*/

class DSO_CLASS Kanji : public Object
{
public:
	enum { SJIS, EUC, JIS };

	Kanji(int _code=SJIS);
	Kanji(const std::string& filename, int _code=SJIS);

	void print(int x, int y, const char *str);

	void setCode(int _code);
	int getFontSize();
	bool load(const std::string& filename);

protected:
	void init();
	const unsigned char* getChar(int index);

	int font_size;
	int char_byte;
	int max_index;
	int code;
	std::vector<unsigned char> data;

};




}	// ist




#endif
