#ifdef _WIN32
	#include <windows.h>
#else
	#include <dlfcn.h>
	#include <unistd.h>
	#include <sys/stat.h>
	#include <sys/types.h>
	#include <dirent.h>
#endif
#include <cstdio>
#include <zlib.h>
#include "ist_sys.h"



#define IGZ_VERSION 100

namespace ist {

/*
	Directory
*/

bool IsDir(const char *path) {
#ifdef _WIN32
	return (::GetFileAttributes(path)&FILE_ATTRIBUTE_DIRECTORY)!=0;
#else
	struct stat st;
	::stat(path, &st);
	return S_ISDIR(st.st_mode);
#endif
}

bool MakeDir(const char *path) {
#ifdef _WIN32
	return ::CreateDirectory(path, NULL)==TRUE;
#else
	return ::mkdir(path, 0777)==0;
#endif
}

size_t MakeDeepDir(const char *path) {
	if(!path)
		return 0;

	size_t num = 0;
	for(size_t i=0; i<strlen(path); ++i) {
		if(path[i]=='/') {
			std::string p(path, path+i);
			if(ist::MakeDir(p.c_str()))
				++num;
		}
	}
	return num;
}

bool RemoveDir(const char *path) {
#ifdef _WIN32
	return ::RemoveDirectory(path)==TRUE;
#else
	return ::rmdir(path)==0;
#endif
}

bool Remove(const char *path) {
	if(IsDir(path))
		return RemoveDir(path);
	else
		return std::remove(path)==0;
}

bool RemoveRecursive(const char *path) {
	bool res = false;
	if(IsDir(path)) {
		Dir dir(path);
		for(Dir::iterator p=dir.begin(); p!=dir.end(); ++p) {
			if(*p=="." || *p=="..")
				continue;
			std::string next = dir.getPath()+*p;
			res = RemoveRecursive(next.c_str());
		}
		res = RemoveDir(path);
	}
	else {
		res = std::remove(path)==0;
	}
	return res;
}

Dir::Dir() {}
Dir::Dir(const char *path) { open(path); }
size_t Dir::size() const { return m_files.size(); }
const std::string& Dir::operator[](size_t i) const { return m_files[i]; }
const std::string& Dir::getPath() const { return m_path; }

bool Dir::open(const char *path) {
	m_files.clear();
	m_path = path;
	if(m_path[m_path.size()-1]!='/')
		m_path+='/';

#ifdef _WIN32
	WIN32_FIND_DATA wfdata;
	HANDLE handle = ::FindFirstFile((m_path+"*").c_str(), &wfdata);
	if(handle!=INVALID_HANDLE_VALUE) {
		do {
			m_files.push_back(wfdata.cFileName);
		} while(::FindNextFile(handle, &wfdata));
		::FindClose(handle);
	}
	else {
		return false;
	}
	return true;
#else
	DIR *dir = ::opendir(m_path.c_str());
	if(dir!=0) {
		dirent *dr;
		while((dr=::readdir(dir))!=0) {
			m_files.push_back(dr->d_name);
		}
		::closedir(dir);
	}
	else {
		return false;
	}
	return true;
#endif
}

bool Dir::openRecursive(const char *path) {
	Dir dir(path);
	for(iterator p=dir.begin(); p!=dir.end(); ++p) {
		if(*p=="." || *p=="..")
			continue;
		std::string next = dir.getPath()+*p;
		if(IsDir(next.c_str()))
			openRecursive(next.c_str());
		else
			m_files.push_back(next);
	}
	return size()!=0;
}






/*
	DynamicSharedObject
*/

#ifdef _WIN32

DSO::DSO() :
	handle(NULL)
{}

DSO::DSO(const std::string& filename) :
	handle(NULL)
{
	load(filename);
}

DSO::~DSO()
{
	unload();
}

bool DSO::load(const std::string& _filename)
{
	unload();
	filename = _filename;
	handle = ::LoadLibrary(filename.c_str());
	if(!(*this))
		filename.clear();
	return !!(*this);
}

void DSO::unload()
{
	if(!!(*this)) {
		::FreeLibrary(handle);
		filename.clear();
		handle = NULL;
	}
}

void* DSO::getFunction(const std::string& funcname) const
{
	if(!(*this))
		return 0;
	return (void*)::GetProcAddress(handle, funcname.c_str());
}

bool DSO::operator!() const
{
	return handle==NULL;
}

#else // _WIN32


DSO::DSO() : handle(0)
{}

DSO::DSO(const std::string& path) : handle(0)
{
	load(path);
}

DSO::~DSO()
{
	unload();
}

bool DSO::load(const std::string& path)
{
	unload();

	if(path[0]!='/') {
		char buf[128];
		filename = std::string(::getcwd(buf, 128))+'/'+path;
	}
	else
		filename = path;

	handle = ::dlopen(filename.c_str(), RTLD_NOW | RTLD_GLOBAL);

	if(!(*this))
		filename.clear();
	return !!(*this);
}

void DSO::unload()
{
	if(!!(*this)) {
		::dlclose(handle);
		filename.clear();
		handle = 0;
	}
}

void* DSO::getFunction(const std::string& funcname) const
{
	if(!(*this))
		return 0;
	return ::dlsym(handle, funcname.c_str());
}

bool DSO::operator!() const
{
	return handle==0;
}
#endif // _WIN32



/*
	IGZ
*/

IGZCompresser::IGZCompresser(ist::bstream& s) : bf(s) {
	char tag[10];
	memset(tag, 0, 10);
	sprintf(tag, "IGZ %d", IGZ_VERSION);
	bf.write(tag, 10);
}

IGZCompresser::~IGZCompresser() { }

size_t IGZCompresser::add(const char *path) {
	size_t ret = 0;
	if(ist::IsDir(path)) {
		Dir dir(path);
		for(size_t i=0; i<dir.size(); ++i) {
			if(dir[i]!="." && dir[i]!="..")
				ret+=add((dir.getPath()+dir[i]).c_str());
		}
	}
	else {
		ist::bbuffer f;
		if(f.load(path) && addBuffer(path, f))
			++ret;
	}

	return ret;
}

bool IGZCompresser::addBuffer(const char *path, const ist::bbuffer& buf) {
	unsigned long cmpsize = buf.size()*2+12;
	ist::bbuffer cmpbuf;
	cmpbuf.resize(cmpsize);
	compress((Bytef*)cmpbuf.ptr(), &cmpsize, (Bytef*)buf.ptr(), buf.size());

	bf << std::string(path) << buf.size() << size_t(cmpsize);
	bf.write(cmpbuf.ptr(), cmpsize);

	return true;
}


IGZExtracter::IGZExtracter(ist::bstream& s) : bf(s) {
	size_t filesize = bf.size();

	char tag[10];
	int version;
	bf.read(tag, 10);
	if(sscanf(tag, "IGZ %d", &version)!=1)
		return;

	while(!bf.eof()) {
		std::string path;
		size_t size, cmpsize;
		bf >> path;
		entry[path] = bf.rtell();
		bf >> size >> cmpsize;
		if(bf.rtell()+cmpsize>=filesize)
			break;
		bf.rskip(cmpsize);
	}
}

IGZExtracter::~IGZExtracter() { }

bool IGZExtracter::extractBuffer(const std::string& path, ist::bbuffer& buf) {
	entry_cont::iterator p = entry.find(path);
	if(p==entry.end())
		return false;

	bf.rseek(p->second);

	size_t size, cmpsize;
	unsigned long _size;
	ist::bbuffer cmpbuf;

	bf >> size >> cmpsize;
	buf.resize(size);
	cmpbuf.resize(cmpsize);
	bf.read(cmpbuf.ptr(), cmpsize);

	_size = size;
	uncompress((Bytef*)buf.ptr(), &_size, (Bytef*)cmpbuf.ptr(), cmpsize);

	return true;
}

bool IGZExtracter::extractFile(const std::string& path) {
	ist::bbuffer buf;
	if(extractBuffer(path, buf)) {
		MakeDeepDir(path.c_str());
		std::ofstream out(path.c_str(), std::ios::binary);
		if(!!out) {
			out.write(buf.ptr(), buf.size());
			return true;
		}
	}
	return false;
}

size_t IGZExtracter::extractAllFiles() {
	size_t ret = 0;
	for(entry_cont::iterator p=entry.begin(); p!=entry.end(); ++p) {
		if(extractFile(p->first))
			++ret;
	}

	return ret;
}



}

