#ifndef IST_ITERATOR_H
#define IST_ITERATOR_H

#include <boost/shared_ptr.hpp>
#include <vector>
#include <map>

namespace ist {


struct non_const_tag {};
struct const_tag {};



// C^[tF[X 

template<class T>
class IIterator
{
public:
	typedef T value_type;

	virtual ~IIterator() {}
	virtual void reset()=0;
	virtual bool hasNext()=0;
	virtual value_type& iterate()=0;
	virtual value_type& get()=0;
};

template<class T>
class IConstIterator
{
public:
	typedef T value_type;

	virtual ~IConstIterator() {}
	virtual void reset()=0;
	virtual bool hasNext()=0;
	virtual const value_type& iterate()=0;
	virtual const value_type& get()=0;
};



namespace impl {

	// ConstTagconst_tagw肵const TԂ 
	template<class T, class ConstTag>
	struct as_const {
		typedef T type;
	};

	template<class T>
	struct as_const<T, const_tag> {
		typedef const T type;
	};


	// ConstTagconst_tagw肵const_iteratorԂ 
	template<class C, class ConstTag>
	struct as_const_iterator {
		typedef typename C::iterator type;
	};

	template<class C>
	struct as_const_iterator<C, const_tag> {
		typedef typename C::const_iterator type;
	};




	// STLRei^
	template <
		class Container,
		class ConstTag = non_const_tag
	>
	class Position
	{
	public:
		typedef typename as_const<Container, ConstTag>::type container;
		typedef typename as_const_iterator<Container, ConstTag>::type iterator;

		Position() : _cont(0) {}

		void reset() {
			if(_cont)
				reset(_cont->begin(), _cont->end());
		}

		void reset(container& c) {
			_cont = &c;
			reset();
		}

		void reset(iterator b, iterator e) {
			_begin = b;
			_end = e;
			_it = _begin;
		}

		bool hasNext() {
			if(_it!=_end)
				return true;
			reset();
			return false;
		}

		iterator _iterate() {
			return _it++;
		}

		iterator _get() {
			return _it;
		}

		container& getContainer() { return *_cont; }
		const container& getContainer() const { return *_cont; }
		iterator begin() { return _begin; }
		iterator end() { return _end; }
		iterator current() { return _it; }

	private:
		container *_cont;
		iterator _begin;
		iterator _end;
		iterator _it;
	};


	template <class Container, class ConstTag = non_const_tag>
	class ContainerPosition : public Position<Container, ConstTag>
	{
	public:
		typedef typename as_const<Container, ConstTag>::type container;
		typedef typename as_const<typename Container::value_type, ConstTag>::type value_type;
		ContainerPosition() {}
		ContainerPosition(container& c) { this->reset(c); }
		value_type& iterate() { return *this->_iterate(); }
		value_type& get() { return *this->_get(); }
	};


	template <class Container, class ConstTag = non_const_tag>
	class MapContainerPosition : public Position<Container, ConstTag>
	{
	public:
		typedef typename as_const<Container, ConstTag>::type container;
		typedef typename as_const<typename Container::mapped_type, ConstTag>::type value_type;
		MapContainerPosition() {}
		MapContainerPosition(container& c) { this->reset(c); }
		value_type& iterate() { return this->_iterate()->second; }
		value_type& get() { return this->_get()->second; }
	};



	template<class ValueType, class ConstTag>
	class IteratorType : public IIterator<ValueType>
	{
	public:
		typedef IIterator<ValueType> Interface;
	};

	template<class ValueType>
	class IteratorType<ValueType, const_tag> : public IConstIterator<ValueType>
	{
	public:
		typedef IConstIterator<ValueType> Interface;
	};
} // impl 




// ContainerIterator 

template<
	class Container,
	class ConstTag = non_const_tag,
	class ValueType = typename Container::value_type,
	class Position = impl::ContainerPosition<Container, ConstTag>
>
class ContainerIterator : public impl::IteratorType<ValueType, ConstTag>
{
public:
	typedef ConstTag const_t;
	typedef typename impl::as_const<Container, ConstTag>::type container;
	typedef Position pos_t;
	typedef typename pos_t::iterator iterator;
	typedef typename pos_t::value_type value_type;

	void reset() { pos.reset(); }
	bool hasNext() { return pos.hasNext(); }
	value_type& iterate() { return pos.iterate(); }
	value_type& get() { return pos.get(); }

	ContainerIterator() {}
	ContainerIterator(container& c) { reset(c); }
	void reset(container& c) { pos.reset(c); }
	void reset(iterator b, iterator e) { pos.reset(b, e); }
	iterator begin() { return pos.begin(); }
	iterator end() { return pos.end(); }
	iterator current() { return pos.current(); }

protected:
	pos_t pos;
};


// MapContainerIterator 

template<
	class Container,
	class ConstTag = non_const_tag,
	class ValueType = typename Container::mapped_type,
	class Position = impl::MapContainerPosition<Container, ConstTag>
>
class MapContainerIterator : public ContainerIterator<Container, ConstTag, ValueType, Position>
{
public:
	typedef typename impl::as_const<Container, ConstTag>::type container;

	MapContainerIterator() {}
	MapContainerIterator(container& c) { this->reset(c); }
};





// MultiContainerIterator 
// ^CṽRei𕡐lߍ߂郂m 

template <
	class Container,
	class ConstTag = non_const_tag,
	class ValueType = typename Container::value_type,
	class Position = impl::ContainerPosition<Container, ConstTag>
>
class MultiContainerIterator : public impl::IteratorType<ValueType, ConstTag>
{
public:
	typedef ConstTag const_t;
	typedef typename impl::as_const<Container, ConstTag>::type container;
	typedef Position pos_t;
	typedef typename pos_t::value_type value_type;
	typedef std::vector<pos_t> pos_cont;
	typedef typename pos_cont::iterator iterator;

public:
	void reset() {
		for(size_t i=0; i<poses.size(); ++i)
			poses[i].reset();
		current = 0;
	}

	bool hasNext() {
		while(current<poses.size()) {
			if(poses[current].hasNext())
				return true;
			++current;
		}
		current = 0;
		return false;
	}

	value_type& iterate() {
		return poses[current].iterate();
	}

	value_type& get() {
		return poses[current].get();
	}


	MultiContainerIterator() : current(0) {}

	void add(container& c) {
		poses.push_back(pos_t(c));
	}

	void add(const pos_t& p) {
		poses.push_back(p);
		poses.back().reset();
	}

	void clear() {
		poses.clear();
		current = 0;
	}

	size_t size() const { return poses.size(); }
	container& operator[](size_t i) { return poses[i].getContainer(); }
	iterator begin() { return poses.begin(); }
	iterator end() { return poses.end(); }

private:
	pos_cont poses;
	size_t current;
};



// Fȃ^CṽReilߍ߂郂m 
template <
	class T,
	class ConstTag = non_const_tag
>
class MultiTypeIterator : public impl::IteratorType<T, ConstTag>
{
public:
	typedef typename impl::IteratorType<T, ConstTag>::Interface Interface;
	typedef T value_type;
	typedef boost::shared_ptr<Interface> i_ptr;
	typedef std::vector<i_ptr> i_ptr_cont;
	typedef typename i_ptr_cont::iterator iterator;

public:
	void reset() {
		for(size_t i=0; i<its.size(); ++i)
			its[i]->reset();
		current = 0;
	}

	bool hasNext() {
		while(current<its.size()) {
			if(its[current]->hasNext())
				return true;
			++current;
		}
		current = 0;
		return false;
	}

	value_type& iterate() {
		return its[current]->iterate();
	}

	value_type& get() {
		return its[current]->get();
	}



	MultiTypeIterator() : current(0) {}

	void add(Interface *p) {
		its.push_back(i_ptr(p));
	}

	template<class Iterator>
	void addIterator(const Iterator& p) {
		its.push_back(i_ptr(new Iterator(p)));
	}

	template<class Container>
	void addContainer(Container& c) {
		its.push_back(i_ptr(new ContainerIterator<Container, ConstTag>(c)));
	}

	template<class Container>
	void addMapContainer(Container& c) {
		its.push_back(i_ptr(new MapContainerIterator<Container, ConstTag>(c)));
	}

	void clear() {
		its.clear();
		current = 0;
	}

	size_t size() { return its.size(); }
	Interface& operator[](size_t i) { return *its[i]; }
	iterator begin() { return its.begin(); }
	iterator end() { return its.end(); }

private:
	i_ptr_cont its;
	size_t current;
};


} // namespace ist

#endif // IST_ITERATOR_H
