#include <string>
#include <map>
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include "ist/dso_setup.h"

class IGame;
extern IGame& _GetIGame();
#define GetIGame() _GetIGame()
#include "aos/aos.h"



class BGM : public ist::Object
{
public:
	BGM() : music(0) {}
	BGM(const std::string& filename) : music(0) { load(filename); }

	~BGM() {
		if(music) {
			halt();
			Mix_FreeMusic(music);
		}
	}

	bool load(const std::string& filename) {
		if(music)
			Mix_FreeMusic(music);
		music = Mix_LoadMUS(filename.c_str());
		return music!=0;
	}

	bool play() {
		if(!music)
			return false;
		Mix_PlayMusic(music, -1);
		return true;
	}

	bool operator!() {
		return music==0;
	}

	static void halt() {
		Mix_HaltMusic();
	}

private:
	Mix_Music *music;
};


class Sound : public ist::Object
{
public:
	Sound() : channel(0), chunk(0) {}
	Sound(const std::string& filename) : channel(0), chunk(0) { load(filename); }

	~Sound() {
		if(chunk) {
			Mix_FreeChunk(chunk);
		}
	}

	bool load(const std::string& filename) {
		if(chunk)
			Mix_FreeChunk(chunk);

		chunk = Mix_LoadWAV(filename.c_str());
		return chunk!=0;
	}

	bool play(int c) {
		if(!chunk)
			return false;
		if(Mix_Playing(c)) {
			return false;
		//	Mix_HaltChannel(c);
		}
		Mix_PlayChannel(c, chunk, 0);
	//	channel = Mix_PlayChannel(-1, chunk, 0);
		return true;
	}

	bool operator!() {
		return chunk==0;
	}

	void halt() {
		Mix_HaltChannel(channel);
	}

private:
	int channel;
	Mix_Chunk *chunk;
};




class DummySoundPlayer : public ISoundPlayer
{
public:
	virtual bool loadSE(const std::string& filename) { return false; }
	virtual bool playSE(const std::string& filename, int c) { return false; }

	virtual bool loadBGM(const std::string& fiename) { return false; }
	virtual bool playBGM(const std::string& fiename) { return false; }
	virtual void haltBGM() {}
};


class SoundPlayer : public ISoundPlayer
{
public:
	virtual bool loadSE(const std::string& filename) {
		sound_ptr& s = sound[filename];
		if(!s)
			s = new Sound();
		return s->load(filename);
	}

	virtual bool playSE(const std::string& filename, int c) {
		if(Mix_Playing(c))
			return false;

		sound_ptr& s = sound[filename];
		if(!s)
			s = new Sound(filename);
		if(!!(*s))
			return s->play(c);
		return false;
	}

	virtual void haltSE() {
		Mix_HaltChannel(-1);
	}



	virtual bool loadBGM(const std::string& filename) {
		bgm_ptr& s = bgm[filename];
		if(!s)
			s = new BGM();
		return s->load(filename);
	}

	virtual bool playBGM(const std::string& filename) {
		bgm_ptr& s = bgm[filename];
		if(!s)
			s = new BGM(filename);
		if(!!(*s)) {
			s->play();
			return true;;
		}
		return false;
	}

	virtual void haltBGM() {
		Mix_HaltMusic();
	}

private:
	typedef boost::intrusive_ptr<Sound> sound_ptr;
	typedef boost::intrusive_ptr<BGM>   bgm_ptr;
	typedef std::map<std::string, sound_ptr> sound_cont;
	typedef std::map<std::string, bgm_ptr>   bgm_cont;
	sound_cont sound;
	bgm_cont   bgm;
};





namespace {
	IGame *igame=0;
}
IGame& _GetIGame() { return *igame; }

DECLARE_DSO_ENTRYPOINT();


EXPORT void OnLoad(IGame& ig)
{
	igame= &ig;
	if(Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 2, 1024)<0)
		std::cout << "SDL_mixeȑɎs܂B" << std::endl;
	else
		GetIGame().setSoundPlayer(new SoundPlayer());
}


EXPORT void OnRelease()
{
	Mix_CloseAudio();
}

