#include <queue>
#include <stdarg.h>
#include <SDL/SDL_thread.h>
#include <zlib.h>
#include "ist/bstream.h"
#include "game.h"
#include "event_router.h"







typedef boost::shared_ptr<ist::TCPSocket> socket_ptr;

class Mutex
{
public:
	Mutex() {
		m_mutex = SDL_CreateMutex();
	}

	~Mutex() {
		SDL_DestroyMutex(m_mutex);
	}

	void lock() {
		SDL_LockMutex(m_mutex);
	}

	void unlock() {
		SDL_UnlockMutex(m_mutex);
	}

private:
	SDL_mutex *m_mutex;

	const Mutex& operator=(const Mutex&); // Rs[s 
	Mutex(const Mutex&);
};

class ScopedLock
{
public:
	ScopedLock(Mutex& m) : m_mutex(m) {
		m_mutex.lock();
	}
	~ScopedLock() {
		m_mutex.unlock();
	}

private:
	Mutex& m_mutex;
};



enum {
	ANY_EMPTY,
	ANY_INT,
	ANY_SIZE_T,
	ANY_FLOAT,
	ANY_BOOL,
	ANY_STRING,
	ANY_VECTOR4,
};

ist::bstream& operator<<(ist::bstream& l, boost::any& a) {
	if(a.empty()) {
		l << ANY_EMPTY;
	}
	else if(a.type()==typeid(int)) {
		l << ANY_INT;
		l << ToInt(a);
	}
	else if(a.type()==typeid(gobj_id)) {
		l << ANY_SIZE_T;
		l << ToSizeT(a);
	}
	else if(a.type()==typeid(float)) {
		l << ANY_FLOAT;
		l << ToFloat(a);
	}
	else if(a.type()==typeid(bool)) {
		l << ANY_BOOL;
		l << ToBool(a);
	}
	else if(a.type()==typeid(std::string)) {
		l << ANY_STRING;
		l << ToString(a);
	}
	else if(a.type()==typeid(ist::vector4)) {
		l << ANY_VECTOR4;
		l << ToVector4(a);
	}
	else {
		printf("serialize_any: unknown type detected.\n");
	}
	return l;
}

ist::bstream& operator>>(ist::bstream& l, boost::any& a) {
	int type;
	l >> type;
	if(type==ANY_EMPTY) {
	}
	else if(type==ANY_INT) {
		int v;
		l >> v;
		a = v;
	}
	else if(type==ANY_SIZE_T) {
		size_t v;
		l >> v;
		a = v;
	}
	else if(type==ANY_FLOAT) {
		float v;
		l >> v;
		a = v;
	}
	else if(type==ANY_BOOL) {
		bool v;
		l >> v;
		a = v;
	}
	else if(type==ANY_STRING) {
		std::string v;
		l >> v;
		a = v;
	}
	else if(type==ANY_VECTOR4) {
		ist::vector4 v;
		l >> v;
		a = v;
	}
	else {
		printf("deserialize_any: unknown type detected\n");
	}

	return l;
}




void writeEvent(ist::bstream& l, const Event& e) {
	l << e.type;
	l << e.frame;

	if(e.getType()==EVT_DESTROY) {
		l << e.from;
		l << e.to;
	}
	else if(e.getType()==EVT_ACTION) {
		l << e.from;
		l << e.to;

		Event::ActionInfo& info = e.getActionInfo();
		gobj_id id =  info.target_obj ? info.target_obj->getID() : info.target_id;
		l << info.target_pos;
		l << id;
		l << info.flag;
	}
	else if(e.getType()==EVT_COMMAND) {
		l << e.from;
		l << e.to;

		Event::CommandInfo& info = e.getCommandInfo();
		l << info.name;
		l << info.value;
	}
	else if(e.getType()==EVT_CALL) {
		l << e.from;
		l << e.to;

		Event::CallInfo& info = e.getCallInfo();
		l << info.name;
		l << info.value;
	}
	else if(e.getType()==EVT_SYSTEM) {
		Event::SystemInfo& info = e.getSystemInfo();
		l << info.name;
		l << info.value;
	}
	else if(e.getType()==EVT_CREATE) {
		Event::CreateInfo& info = e.getCreateInfo();
		l << info.type;
		l << info.id;
		if(info.type!=UNITGROUP) {
			l << info.classname;
			l << info.faction;
		}
		else {
			l << info.group.size();
			for(size_t i=0; i<info.group.size(); ++i)
				l << info.group[i];
		}
	}
	else {
		printf("writeEvent: unknown event detected %d\n", e.getType());
	}
}

bool readEvent(ist::bstream& l, Event& e) {
	if(!l)
		return false;

	l >> e.type;
	l >> e.frame;
	if(e.getType()==EVT_DESTROY) {
		l >> e.from;
		l >> e.to;
	}
	else if(e.getType()==EVT_ACTION) {
		l >> e.from;
		l >> e.to;

		e.info = new Event::ActionInfo(ist::vector4(), 0, 0, 0);
		Event::ActionInfo& info = e.getActionInfo();
		l >> info.target_pos;
		l >> info.target_id;
		l >> info.flag;
	}
	else if(e.getType()==EVT_COMMAND) {
		l >> e.from;
		l >> e.to;

		e.info = new Event::CommandInfo(std::string(), boost::any());
		Event::CommandInfo& info = e.getCommandInfo();
		l >> info.name;
		l >> info.value;
	}
	else if(e.getType()==EVT_CALL) {
		l >> e.from;
		l >> e.to;

		e.info = new Event::CallInfo(std::string(), boost::any());
		Event::CallInfo& info = e.getCallInfo();
		l >> info.name;
		l >> info.value;
	}
	else if(e.getType()==EVT_SYSTEM) {
		e.info = new Event::SystemInfo(std::string(), boost::any());
		Event::SystemInfo& info = e.getSystemInfo();
		l >> info.name;
		l >> info.value;
	}
	else if(e.getType()==EVT_CREATE) {
		e.info = new Event::CreateInfo();
		Event::CreateInfo& info = e.getCreateInfo();
		l >> info.type;
		l >> info.id;
		if(info.type!=UNITGROUP) {
			l >> info.classname;
			l >> info.faction;
		}
		else {
			size_t s;
			l >> s;
			for(size_t i=0; i<s; ++i) {
				info.group.push_back(0);
				l >> info.group.back();
			}
		}
	}
	else {
		printf("readEvent: unknown event detected %d\n", e.getType());
		return false;
	}
	return true;
}





void SequentialWrite(ist::TCPSocket& sst, ist::bbuffer& v) {
	ist::bbuffer b;
	char buf[64];

	sprintf(buf, "ET: %d\r\n", v.size());
	b.write(buf, strlen(buf));
	b.write(v.ptr(), v.size());
	sst.send(b.ptr(), b.size());

	v.clear();
}

bool SequentialRead(ist::TCPSocket& sst, ist::bbuffer& v) {
	static std::string head;
	size_t len;

	head.clear();
	v.clear();
	std::getline(sst, head);
	if(sscanf(head.c_str(), "ET: %d", &len)==1) {
		v.resize(len);
		sst.read(v.ptr(), len);
		return true;
	}
	return false;
}



class EventQueue
{
public:
	typedef std::list<Event> event_cont;
	typedef event_cont::iterator iterator;

	void pushN(const Event& e) {
		evt_queue.push_back(e);
	}

	void popN() {
		if(!evt_queue.empty())
			evt_queue.pop_front();
	}

	void clearN() {
		evt_queue.clear();
	}

	// sync
	void pushS(const Event& e) {
		ScopedLock lock(mutex);
		pushN(e);
	}

	// sync
	void popS() {
		ScopedLock lock(mutex);
		popN();
	}

	// sync
	void clearS() {
		ScopedLock lock(mutex);
		clearN();
	}

	Event& front()   { return evt_queue.front(); }
	bool empty()     { return evt_queue.empty(); }
	Mutex& getMutex(){ return mutex; }

	void erase(iterator p) { evt_queue.erase(p); }
	iterator begin() { return evt_queue.begin(); }
	iterator end()   { return evt_queue.end(); }

private:
	Mutex mutex;
	event_cont evt_queue;
};





class SimpleEventRouter : public IEventRouter
{
public:
	SimpleEventRouter() {}
	~SimpleEventRouter() {}
	int getType() const { return SIMPLE; }

	bool load(const std::string& filename) { return false; }

	void send(const Event& e) {
		if(!fetchCreateEvent(e) && !fetchCallEvent(e) && !fetchSystemEvent(e) && !fetchLocalEvent(e)) {
			if(gobj_ptr p = GetGObj(e.to))
				p->sendEvent(e);
		}
	}

	void update() {}
	void draw() const {}
	bool conditionToStart() const { return true; }



	virtual gobj_ptr fetchCreateEvent(const Event& e) {
		if(e.getType()==EVT_CREATE) {
			Event::CreateInfo& info = e.getCreateInfo();
			gobj_ptr p;
			if(info.type==FACTION)
				p = GetIGame().createFaction(info.classname);
			else if(info.type==UNIT && !!info.faction)
				p = GetIGame().createUnit(info.classname, ToFaction(GetGObj(info.faction)));
			else if(info.type==BUILDING && !!info.faction)
				p = GetIGame().createBuilding(info.classname, ToFaction(GetGObj(info.faction)));
			else if(info.type==MATERIAL)
				p = GetIGame().createMaterial(info.classname);
			else if(info.type==UNITGROUP) {
				IDIteratorAdapter<Event::CreateInfo::id_cont> iter(info.group.begin(), info.group.end());
				p = GetIGame().createGroup(iter);
			}
			if(!p) {
				printf("fetchCreateEvent: %s\n", info.classname.c_str());
				fflush(stdout);
				assert(0);
			}
			p->sendEvent(e);
			p->update();
			AddGObj(p);
			return p;
		}
		return gobj_ptr();
	}

	virtual bool fetchCallEvent(const Event& e) {
		if(e.getType()==EVT_CALL) {
			Event::CallInfo& info = e.getCallInfo();
			if(gobj_ptr p = GetGObj(e.to))
				p->call(info.name, info.value);
			return true;
		}
		return false;
	}

	virtual bool fetchLocalEvent(const Event& e) {
		if(e.getType()==EVT_SYSTEM) {
			Event::SystemInfo& info = e.getSystemInfo();
			if(info.name=="setPlayerFaction") {
				if(faction_ptr f = ToFaction(ToGObj(info.value)))
					SetPlayerFaction(f);
				return true;
			}
		}
		return false;
	}

	virtual bool fetchSystemEvent(const Event& e) {
		bool res = true;
		if(e.getType()==EVT_SYSTEM) {
			Event::SystemInfo& info = e.getSystemInfo();
			if(info.name=="enableSight")
				GetCurrentWorld().enableSight(ToBool(info.value));
			else if(info.name=="loadGround")
				GetCurrentWorld().loadGround(ToString(info.value));
			else if(info.name=="loadGroundTexture")
				GetCurrentWorld().loadGroundTexture(ToString(info.value));
			else if(info.name=="enableGroundLighting")
				GetCurrentWorld().enableGroundLighting(ToBool(info.value));
			else if(info.name=="enableGroundFog")
				GetCurrentWorld().enableGroundFog(ToBool(info.value));
			else {
				printf("unknown system event: %s\n", info.name.c_str());
				res = false;
			}
		}
		else {
			res = false;
		}
		return res;
	}
};

class EventRouter : public SimpleEventRouter
{
public:
	EventRouter() {
		char fname[64];
		size_t t = ::time(0);
		sprintf(fname, "savegame/%d.evt", t);
		f = gzopen(fname, "wb");
	}
	~EventRouter() { gzclose(f); }
	int getType() const { return SINGLE; }

	void send(const Event& e) {
		record(e);
		SimpleEventRouter::send(e);
	}

	virtual void record(const Event& e) {
		ist::bbuffer buf;
		writeEvent(buf, e);
		if(buf.size()>4)
			gzwrite(f, buf.ptr(), buf.size());
	//	std::string l;
	//	fprintf(f, "%s", event_to_string(l, e).c_str());
	}

private:
	gzFile f;
};










class NetEventRouter : public EventRouter
{
public:
	NetEventRouter() : cstart(false) {}
	bool conditionToStart() const { return cstart; }


	bool cstart;
};


class ServerEventRouter : public NetEventRouter
{
public:
	class Client;
	typedef boost::shared_ptr<Client> client_ptr;
	typedef std::map<size_t, client_ptr> client_cont;
	typedef std::list<Event> event_queue;


	class Client
	{
	public:
		ServerEventRouter *er;
		socket_ptr sst;
		size_t id;
		ist::bbuffer sender;
		ist::bbuffer reciever;
		Mutex mutex_send;
		Mutex mutex_recv;
		std::string name;
		std::string status_text;
		volatile frame_t frame;

		size_t ping;
		size_t time_last;
		size_t rps;
		size_t crps;
		size_t sec_last;

		Client(ServerEventRouter *e, ist::TCPSocket *s) : er(e), sst(s), id(0), frame(0),
			ping(0), time_last(0), rps(0), crps(0), sec_last(0) {
		}

		ServerEventRouter& getRoot() { return *er; }
		ist::TCPSocket& getSocket() { return *sst; }
		Mutex& getSendMutex() { return mutex_send; }


		void updatePing() {
			size_t t = SDL_GetTicks();
			ping = t-time_last;
			time_last = t;

			++crps;
			if(t-sec_last>1000) {
				rps = crps;
				crps = 0;
				sec_last = t;
			}
		}
	};


	// thread
	static int waitClient(void *arg) {
		client_ptr cli((Client*)arg);
		ServerEventRouter& er = cli->getRoot();
		ist::TCPSocket& sst = cli->getSocket();
		ist::bbuffer tmp;
		ist::bbuffer out;

		bool accepted = false;
		while(!!sst && !er.end_flag && sst.select(5)) {
			int code = 0;
			SequentialRead(sst, tmp);
			tmp >> code;
			if(code==EVT_LOGIN) {
				er.insertClient(cli);
				accepted = true;

				tmp >> cli->name; // oH 
				out << int(EVT_LOGIN_ACCEPT) << cli->id;
				SequentialWrite(sst, out);
				break;
			}
			else if(code==EVT_PING) {
				out << int(EVT_PONG);
				SequentialWrite(sst, out);
			}
			else if(code==EVT_QUERY_VERSION) {
				std::string i = std::string(VERSION_STRING);
				out << int(EVT_VERSION);
				out.write(i.c_str(), i.size());
				SequentialWrite(sst, out);
			}
			else if(code==EVT_QUERY_PLAYERS) {
				std::string i = er.getPlayersInfo();
				out << int(EVT_PLAYERS);
				out.write(i.c_str(), i.size());
				SequentialWrite(sst, out);
			}
			else if(code==EVT_DISCONNECT) {
				break;
			}
		}

		while(accepted && !!sst && !er.end_flag) {
		/*
			// ܂10boĂȂ̂ɑ^CAEgۂcH 
			if(!sst.select(10)) {
				std::string cli_name = cli->name; // Rs[S 
				printf("Server: %s Ƃ̒ʐM^CAEg܂\n", cli_name.c_str());

				sst.close();
				break;
			}
		*/
			if(sst.select(3)) {
				SequentialRead(sst, tmp);
				if(!tmp.empty()) {
					ScopedLock lock(cli->mutex_recv); // sync recv
					cli->reciever+=tmp;
				}
				cli->updatePing();
				if(cli->ping==0) // LAN̏ꍇbԒʐM4000ƂɂȂCPU̂ŋx܂ 
					SDL_Delay(1);
			}
			{
				ScopedLock lock(cli->mutex_send); // sync send
				if(cli->frame>0)
					cli->sender << int(EVT_SYNC) << cli->frame;
				SequentialWrite(sst, cli->sender);
			}


			if(!sst) {
				std::string cli_name = cli->name;
				printf("Server: %s ʐM܂\n", cli_name.c_str());
			}
		}
		if(!!sst && er.end_flag) {
			tmp << int(EVT_DISCONNECT);
			SequentialWrite(sst, tmp);
		}
		er.running--;
		er.eraseClient(cli);
		return 0;
	}

	// thread
	static int lobyThread(void *arg) {
		ServerEventRouter& er = *((ServerEventRouter*)arg);
		gimite::server_stream_socket server(GetGameConfig().port);
		bool db_added = false;

		if(!!server) {
			char post[512];
			std::string& pname = GetGameConfig().player_name;
			if(pname.size()>32)
				pname.resize(32);	// vCƃobt@I[o[t[łԂ 
			sprintf(post, "mode=add&name=%s&port=%d", pname.c_str(), GetGameConfig().port);

			ist::HTTP h;
			h.post(GetGameConfig().server_database, post);
			if(!!h)
				db_added = true;
		}

		while(!!server && !er.end_flag && !er.conditionToStart()) {
			gimite::socket_t s = server.accept();
			if(!er.end_flag) {
				er.running++;
				SDL_CreateThread(waitClient, new Client(&er, new ist::TCPSocket(s)));
			}
			else {
				ist::TCPSocket sst(s); //  
			}
		}

		if(db_added) {
			ist::HTTP h;
			h.post(GetGameConfig().server_database, "mode=del");
		}

		er.running--;

		return 0;
	}


	class EnterKeyObserver : public IKeyObserver
	{
	public:
		EnterKeyObserver(ServerEventRouter *e) : er(e) {}
		bool isExclusive() { return true; }
		void exec(int i) {
			if(i==SDL_KEYUP) {
				er->start();
			}
		}

		ServerEventRouter *er;
	};

	EventQueue evt_local;
	EventQueue evt_send;
	client_cont clients;
	keyobserv_ptr enter;
	ist::bbuffer worker;
	volatile int running;
	volatile bool end_flag;
	mutable Mutex mutex_client;


	ServerEventRouter() : running(0), end_flag(false) {
		GetCurrentWorld().setIDSeed(0xf0000000);
		enter = new EnterKeyObserver(this);
		RegistKeyObserver(SDLK_RETURN, enter);

		running++;
		SDL_CreateThread(lobyThread, this);
	}

	~ServerEventRouter() {
		end_flag = true;
		if(running)
			ist::TCPSocket killer("localhost", GetGameConfig().port); // ҂󂯃XbhE 
		while(running>0)
			SDL_Delay(10);
	}

	int getType() const { return SERVER; }

	gobj_ptr fetchCreateEvent(const Event& e) {
		gobj_ptr p = NetEventRouter::fetchCreateEvent(e);
		if(!!p && p->isFaction() && ToFaction(p)->getFactionID()==1) {
			SetPlayerFaction(ToFaction(p));
			record(Event(EVT_SYSTEM, 0, 0, new Event::SystemInfo("setPlayerFaction", p->getID())));
		}
		return p;
	}


	// XbhĂ΂邩Ȃ 
	bool insertClient(client_ptr cli) {
		if(conditionToStart())	// sync client
			return false;

		ScopedLock lock(mutex_client);
		for(size_t i=2; i<12; ++i) {
			if(!clients[i]) {
				cli->id = i;
				clients[i] = cli;
				return true;
			}
		}
		return false;
	}

	// XbhĂ΂邩Ȃ 
	void eraseClient(client_ptr cli) {
		ScopedLock lock(mutex_client);	// sync client
		clients.erase(cli->id);
	}

	// XbhĂ΂邩Ȃ 
	std::string getPlayersInfo() const {
		std::string info;

		info+=GetGameConfig().player_name;
		ScopedLock lock(mutex_client);	// sync client
		for(client_cont::const_iterator p=clients.begin(); p!=clients.end(); ++p) {
			info+=", ";
			info+=p->second->name;
		}
		return info;
	}

	void sendToAllClient(const Event& e) {	// sync client
		ScopedLock lock(mutex_client);
		for(client_cont::iterator p=clients.begin(); p!=clients.end(); ++p) {
			client_ptr& cli = p->second;
			ScopedLock lock(cli->mutex_send); // sync send
			writeEvent(cli->sender, e);
		}
	}


	void start() {
		if(!conditionToStart()) {
			cstart = true;
			Event e(EVT_SYSTEM, 0, 0, new Event::SystemInfo("start", boost::any()));
			sendToAllClient(e);
		}
		DeregistKeyObserver(SDLK_RETURN, enter);
	}


	// local event
	void send(const Event& e) {
		evt_local.pushN(e);
	}


	void update() {
		while(!evt_local.empty()) {
			evt_send.pushN(evt_local.front());
			evt_local.popN();
		}
		{
			ScopedLock lock(mutex_client);	// sync client
			for(client_cont::iterator p=clients.begin(); p!=clients.end(); ++p) {
				client_ptr cli = p->second;
				cli->frame = GetCurrentFrame();
				{
					ScopedLock lock(cli->mutex_recv);	// sync recv
					worker = cli->reciever;
					cli->reciever.clear();
				}
				Event e;
				while(!!worker) {
					if(readEvent(worker, e))
						evt_send.pushN(e);
				}
			}
		}
		if(!evt_send.empty()) {
			for(EventQueue::iterator p=evt_send.begin(); p!=evt_send.end();) {
				Event& e = *p;
				e.frame = GetCurrentFrame();
				record(e);
				if(fetchLocalEvent(e)) {
					evt_send.erase(p++);
					continue;
				}
				if(!fetchCreateEvent(e) && !fetchCallEvent(e) && !fetchSystemEvent(e)) {
					if(gobj_ptr g = GetGObj(e.to)) {
						g->sendEvent(e);
					}
					else {
						evt_send.erase(p++);
						continue;
					}
				}
				++p;
			}

			ScopedLock lock(mutex_client);	// sync client
			for(client_cont::iterator c=clients.begin(); c!=clients.end(); ++c) {
				client_ptr& cli = c->second;
				ScopedLock lock(cli->mutex_send); // sync send
				for(EventQueue::iterator p=evt_send.begin(); p!=evt_send.end(); ++p)
					writeEvent(cli->sender, *p);
			}
			evt_send.clearN();
		}

		if(running==0) {
			RequestSetCurrentWorld("title");
		}
	}

	void draw() const {
		char text[128];
		if(!conditionToStart()) {
			sprintf(text, "EnterŃQ[Jn");
			DrawFont(300, 240, text);
		}

		ScopedLock lock(mutex_client);	// sync client
		int i = 0;
		for(client_cont::const_iterator p=clients.begin(); p!=clients.end(); ++p) {
			++i;
			client_ptr c = p->second;
			sprintf(text, "%s ping:%d rps:%d", c->name.c_str(), c->ping, c->rps);
			DrawFont(500, 10+17*i, text);
		}
	}
};




class ClientEventRouter : public NetEventRouter
{
public:
	size_t id;
	int running;
	size_t begin_tick;
	bool start;
	volatile bool end_flag;
	ist::bbuffer sender;
	ist::bbuffer reciever;
	ist::bbuffer worker;
	Mutex mutex_send;
	Mutex mutex_recv;
	EventQueue evt_recv;
	size_t state;

	struct Info
	{
		std::string message;
		frame_t frame;
		size_t warn_level;

		Info(const std::string& i, frame_t f=0, size_t wl=0) : message(i), frame(f), warn_level(wl) {}
	};
	mutable std::list<Info> info;

	size_t ping;
	size_t time_last;
	size_t rps;
	size_t crps;
	size_t sec_last;

	enum {
		NST_NOT_CONNECTED=100,
		NST_CONNECTING,
		NST_LOGIN,
		NST_OK,
		NST_CONNECT_FAILED,
		NST_LOGIN_FAILED,
		NST_DISCONNECTED,
		NST_SERVERSHUTDOWN,
		NST_TIMEOUT,
	};

	static std::string getNSTMessage(size_t i) {
		std::string mes;
		switch(i) {
		case NST_NOT_CONNECTED: mes = ""; break;
		case NST_CONNECTING:    mes = "T[o[ɐڑc"; break;
		case NST_LOGIN:         mes = "Q[ւ̎Qvc"; break;
		case NST_OK:            mes = "Q[֎Q܂"; break;
		case NST_CONNECT_FAILED:mes = "T[o[ւ̐ڑɎs܂"; break;
		case NST_LOGIN_FAILED:  mes = "Q[ւ̎QɎs܂"; break;
		case NST_DISCONNECTED:  mes = "T[o[R܂"; break;
		case NST_SERVERSHUTDOWN:mes = "T[o[ʐM܂"; break;
		case NST_TIMEOUT:       mes = "T[o[Ƃ̒ʐM^CAEg܂"; break;
		}
		return mes;
	}


	// thread
	static int waitServer(void *arg) {
		ClientEventRouter *er = (ClientEventRouter*)arg;

		er->time_last = ::time(0);
		er->changeNST(NST_CONNECTING);

		ist::bbuffer tmp;
		ist::TCPSocket sst(GetGameConfig().server_address, GetGameConfig().server_port);

		if(!sst)
			er->changeNST(NST_CONNECT_FAILED);
		else
			er->changeNST(NST_LOGIN);

		bool accepted = false;
		if(!!sst) {
			int code = 0;
			tmp << EVT_LOGIN << GetGameConfig().player_name;
			SequentialWrite(sst, tmp);
			if(sst.select(3)) {
				SequentialRead(sst, tmp);
				tmp >> code;
			}
			if(code==EVT_LOGIN_ACCEPT) {
				size_t id = 0;
				tmp >> id;
				er->setClientID(id); // oH 
				accepted = true;
			}
		}

		if(accepted)
			er->changeNST(NST_OK);
		else if(er->state!=NST_CONNECT_FAILED)
			er->changeNST(NST_LOGIN_FAILED);

	//	std::fstream evt_log("evt_log", std::ios::out | std::ios::binary); // fobOp 
		while(accepted && !!sst && !er->end_flag) {
			{
				ScopedLock lock(er->mutex_send);	// sync send
				SequentialWrite(sst, er->sender);
				er->sender.clear();
			}

			if(sst.select(3)) {
				SequentialRead(sst, tmp);
				if(!tmp.empty()) {
				//	evt_log.write(tmp.ptr(), tmp.size()); // fobOp 
					ScopedLock lock(er->mutex_recv);	// sync recv
					er->reciever+=tmp;
				}
				er->updatePing();
			}

			if(!sst)
				er->changeNST(NST_SERVERSHUTDOWN);
		}
		if(!!sst && er->end_flag) {
			tmp << int(EVT_LOGOUT);
			SequentialWrite(sst, tmp);
		}
		er->running--;
		return 0;
	}



	ClientEventRouter() : id(0), running(0), begin_tick(0),
		start(false), end_flag(false), state(NST_NOT_CONNECTED),
		ping(0), time_last(0), rps(0), crps(0), sec_last(0) {
		getNSTMessage(NST_OK);
		printf("Client: %s:%d ɐڑ܂c\n", GetGameConfig().server_address.c_str(), GetGameConfig().server_port);

		running++;
		SDL_CreateThread(waitServer, this);
	}

	~ClientEventRouter() {
		end_flag = true;
		while(running>0)
			SDL_Delay(10);
	}
	int getType() const { return CLIENT; }


	void changeNST(size_t s) {
		state = s;
		printf("Client: %s\n", getNSTMessage(s).c_str());
	}

	void queueMessage(const std::string& mes, size_t wlevel=0) {
		char buf[256];
		sprintf(buf, "[%d] %s\n", GetCurrentFrame(), mes.c_str());
		printf(buf);
		info.push_back(Info(buf, GetCurrentFrame(), wlevel));
		if(info.size()>2)
			info.pop_front();
	}


	void updatePing() {
		size_t t = SDL_GetTicks();
		ping = t-time_last;
		time_last = t;

		++crps;
		if(t-sec_last>1000) {
			rps = crps;
			crps = 0;
			sec_last = t;
		}
	}

	void setClientID(size_t i) {
		id = i;
		GetCurrentWorld().setIDSeed(0x10000000*id);
	}


	gobj_ptr fetchCreateEvent(const Event& e) {
		gobj_ptr p = NetEventRouter::fetchCreateEvent(e);
		if(!!p && p->isFaction() && ToFaction(p)->getFactionID()==size_t(id)) {
			SetPlayerFaction(ToFaction(p));
			record(Event(EVT_SYSTEM, 0, 0, new Event::SystemInfo("setPlayerFaction", p->getID())));
		}
		return p;
	}

	bool fetchServerEvent(const Event& e) {
		bool res = true;

		if(e.getType()==EVT_SYSTEM) {
			Event::SystemInfo& info = e.getSystemInfo();
			if(info.name=="start") {
				start = true;
				begin_tick = SDL_GetTicks();
			}
			else if(info.name=="setFrameLimit") {
				GetCurrentWorld().setFrameLimit(ToInt(info.value));
			}
			else {
				res = false;
			}
		}
		else {
			res = false;
		}
		return res;
	}

	void send(const Event& e) {
		ScopedLock lock(mutex_send);	// sync send
		writeEvent(sender, e);
	}

	void update() {
		if(!cstart) {
			if(start) {
				cstart = true;
			}
		}

		if(!reciever.empty()) {
			ScopedLock lock(mutex_recv);	// sync recv
			worker = reciever;
			reciever.clear();
		}
		Event e;
		while(!!worker) {
			int code(0);
			worker >> code;
			if(code==EVT_SYNC) {
				frame_t f;
				worker >> f;
				GetCurrentWorld().setFrameLimit(f);
			}
			else if(code==EVT_DISCONNECT) {
			//	end_flag = true;
			}
			else {
				worker.rskip(-4);
				if(readEvent(worker, e) && !fetchServerEvent(e)) {
					evt_recv.pushN(e);
				}
			}
		}

	//	if(GetCurrentFrame()%180==0)
	//		queueMessage("ĂƂĂƂ[[DD^}oB", SDL_GetTicks()%3);

		while(!evt_recv.empty()) {
			Event& e = evt_recv.front();
			size_t frame = e.frame;

			if(fetchServerEvent(e)) {
				evt_recv.popN();
			}
			else if(frame==GetCurrentFrame()) {
				record(e);
				if(!fetchCreateEvent(e) && !fetchCallEvent(e) && !fetchSystemEvent(e) && !fetchLocalEvent(e)) {
					if(gobj_ptr p = GetGObj(e.to))
						p->sendEvent(e);
					else {
						queueMessage("݂ȂIuWFNgւ̃CxgmBĂ\܂B", 2);
					}
				}
				evt_recv.popN();
			}
			else if(frame<GetCurrentFrame()) {
				queueMessage("ߋ̃CxgmBĂ\܂B", 2);
				printf("sync error (%d frame)\n", GetCurrentFrame());
			}
			else
				break;
		}

		if(running==0) {
			RequestSetCurrentWorld("title");
		}
	}

	void draw() const {
		char text[128];
		bool connected = false;

		if(state==NST_OK) {
			connected = true;
			sprintf(text, "ping:%d rps:%d", ping, rps);
		}
		else {
			sprintf(text, "%s", getNSTMessage(state).c_str());
		}
		DrawFont(5, 22, text);
		if(connected && !conditionToStart()) {
			sprintf(text, "Q[֎Q܂BJn܂ł҂B");
			DrawFont(5, 22+17, text);
		}

		static ist::Color cols[] = {
			ist::Color(1.0f, 1.0f, 1.0f, 1.0f),
			ist::Color(1.0f, 0.5f, 0.5f, 1.0f),
			ist::Color(1.0f, 0.0f, 0.0f, 1.0f),
		};
		int i=1;
		for(std::list<Info>::iterator p=info.begin(); p!=info.end();) {
			if(GetCurrentFrame()-p->frame>300) {
				info.erase(p++);
			}
			else {
				if(p->warn_level<3)
					cols[p->warn_level].assign();
				DrawFont(150, 25+17*i, p->message.c_str());
				ist::Color().assign();
				++p; ++i;
			}
		}
	}
};


class ReplayEventRouter : public SimpleEventRouter
{
public:
	typedef std::list<Event> event_queue;
	event_queue events;

	bool load(const std::string& filename) {
		ist::bbuffer buf;
		char c;
		gzFile f = gzopen(filename.c_str(), "rb");
		if(!f)
			return false;
		while(gzread(f, &c, 1)!=0)
			buf.push_back(c);
		gzclose(f);

		int count = 0;
		if(!buf.empty()) {
			Event e;
			while(readEvent(buf, e)) {
				events.push_back(e);
				++count;
			}
		}

		return count!=0;
	}

	void send(const Event& e) {}
	void update() {
		while(!events.empty()) {
			Event& e = events.front();

			if(fetchLocalEvent(e)) {
				events.pop_front();
			}
			else if(e.frame==GetCurrentFrame()) {
				if(!fetchCreateEvent(e) && !fetchCallEvent(e) && !fetchSystemEvent(e)) {
					if(gobj_ptr p = GetGObj(e.to))
						p->sendEvent(e);
				}
				events.pop_front();
			}
			else if(e.frame<GetCurrentFrame()) {
				printf("%d frame\n", GetCurrentFrame());
				fflush(stdout);
				assert(0 && "sync error?");
			}
			else
				break;
		}
	}

	void draw() const {
		if(events.empty()) {
			DrawFont(5, 22, "end of replay");
		}
	}
};

erouter_ptr CreateEventRouter(int t) {
	if(t==SIMPLE) return erouter_ptr(new SimpleEventRouter());
	if(t==SINGLE) return erouter_ptr(new EventRouter());
	if(t==SERVER) return erouter_ptr(new ServerEventRouter());
	if(t==CLIENT) return erouter_ptr(new ClientEventRouter());
	if(t==REPLAY) return erouter_ptr(new ReplayEventRouter());
	return erouter_ptr(new EventRouter());
}


