#include "ist/socket.h"
#include "game.h"
#include "ist/bitmap.h"
#include "lua_binder.h"
#include "xml.h"

#ifndef _WIN32
#include <dirent.h>
#endif



Frame::Frame()
{
	if(SDL_Init(SDL_INIT_TIMER | SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK )<0)
		throw std::runtime_error(SDL_GetError());


	const SDL_VideoInfo *info = SDL_GetVideoInfo();
	int bpp = info->vfmt->BitsPerPixel;
	int flag = SDL_OPENGL | GetFrameConfig().ex_flags;

	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
#ifdef _WIN32
	SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 16 );
#endif
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	if( SDL_SetVideoMode( GetFrameConfig().width, GetFrameConfig().height, bpp, flag )==0) {
		std::cout << "damepo" << std::endl;
		throw std::runtime_error(SDL_GetError());
	}

	ist::init();

//	ist::GLInfo glinfo;
//	glinfo.print(std::cout);

	ist::Random::setSeed(::time(0));
}

Frame::~Frame()
{
	ist::quit();
	SDL_Quit();
}







class DummySoundPlayer : public ISoundPlayer
{
public:
	virtual bool loadSE(const std::string&) { return false; }
	virtual bool playSE(const std::string&, int c) { return false; }

	virtual bool loadBGM(const std::string&) { return false; }
	virtual bool playBGM(const std::string&) { return false; }
	virtual void haltBGM() {}
};






class EscObserver : public IKeyObserver
{
public:
	EscObserver() {}
	bool isExclusive() { return true; }
	void exec(int stat) {
		if(stat==SDL_KEYUP) {
			RequestSetCurrentWorld("title");
		}
	}
};




Game::Game() :
	is_shift_key_down(false),
	is_alt_key_down(false),
	is_ctrl_key_down(false),
	end_flag(false)
{
	setCurrentWorld("title");
	sound_player = sound_ptr(new DummySoundPlayer());
}

Game::~Game()
{}




static void PrintScreen()
{
	char filename[128];
	int i=0;
	while(1) {
		sprintf(filename, "sc_%03d.png", i);
		FILE *check=fopen(filename, "r");
		if (check==NULL) {
			break;
		}
		else {
			fclose(check);
			i++;
		}
	}
	
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	std::vector<ist::bRGBA> buf( viewport[2]*viewport[3] );
	glReadBuffer(GL_FRONT);
	glReadPixels(0,0, viewport[2],viewport[3], GL_RGBA, GL_UNSIGNED_BYTE, &buf[0][0] );
	glReadBuffer(GL_BACK);

	ist::Bitmap bmp;
	bmp.resize(viewport[2], viewport[3]);
	for(int i=0; i<viewport[3]; ++i) {
		for(int j=0; j<viewport[2]; ++j) {
			bmp[i][j]=buf[viewport[2]*viewport[3] - (i+1)*viewport[2] + j];
			bmp[i][j].a=255;
		}
	}
	bmp.save(filename);
}



void Game::processEvent()
{
	static bool is_dragging = false;
	static ist::vector2 d_begin;

	SDL_Event e;
	while( SDL_PollEvent(&e) ) {
		switch( e.type ) {
		case SDL_QUIT:
			end_flag = true;
			break;

		case SDL_KEYUP:
		case SDL_KEYDOWN:
			{
				if(e.type==SDL_KEYDOWN) {
					if(e.key.keysym.sym==SDLK_LSHIFT || e.key.keysym.sym==SDLK_RSHIFT)
						is_shift_key_down = true;
					else if(e.key.keysym.sym==SDLK_LALT || e.key.keysym.sym==SDLK_RALT)
						is_alt_key_down = true;
					else if(e.key.keysym.sym==SDLK_LCTRL || e.key.keysym.sym==SDLK_RCTRL)
						is_ctrl_key_down = true;
					else if(e.key.keysym.sym==SDLK_F12)
						PrintScreen();
				}
				else if(e.type==SDL_KEYUP) {
					if(e.key.keysym.sym==SDLK_LSHIFT || e.key.keysym.sym==SDLK_RSHIFT)
						is_shift_key_down = false;
					else if(e.key.keysym.sym==SDLK_LALT || e.key.keysym.sym==SDLK_RALT)
						is_alt_key_down = false;
					else if(e.key.keysym.sym==SDLK_LCTRL || e.key.keysym.sym==SDLK_RCTRL)
						is_ctrl_key_down = false;
					else if(e.key.keysym.sym==SDLK_F11)
						GetGameConfig().nowait = !GetGameConfig().nowait;
				}

				getCurrentWorld().onKey(new Event::KeyInfo(e.key.type, e.key.keysym.sym));
			}
			break;

		case SDL_MOUSEBUTTONUP:
		case SDL_MOUSEBUTTONDOWN:
			{
				if(e.type==SDL_MOUSEBUTTONDOWN && e.button.button==1) {
					is_dragging = true;
					d_begin.x = (float)e.button.x;
					d_begin.y = (float)e.button.y;
				}
				else
					is_dragging = false;
				getCurrentWorld().onButton(
					new Event::ButtonInfo(e.button.type, e.button.button, e.button.x, e.button.y));
			}
			break;

		case SDL_MOUSEMOTION:
			{
				Event::MotionInfo *info = new Event::MotionInfo(is_dragging, e.motion.x, e.motion.y);
				if(is_dragging) {
					ist::PerspectiveCamera& camera = getCurrentWorld().getCamera();
					info->drag_x = d_begin.x;
					info->drag_y = d_begin.y;
					info->dragging_frustom.setup(
						camera.position,
						camera.position + getCurrentWorld().getScreenRay(
							ist::vector2(std::max<float>(d_begin.x, e.motion.x), std::min<float>(d_begin.y, e.motion.y)))*1500.0f,
						camera.position + getCurrentWorld().getScreenRay(
							ist::vector2(std::min<float>(d_begin.x, e.motion.x), std::min<float>(d_begin.y, e.motion.y)))*1500.0f,
						camera.position + getCurrentWorld().getScreenRay(
							ist::vector2(std::min<float>(d_begin.x, e.motion.x), std::max<float>(d_begin.y, e.motion.y)))*1500.0f,
						camera.position + getCurrentWorld().getScreenRay(
							ist::vector2(std::max<float>(d_begin.x, e.motion.x), std::max<float>(d_begin.y, e.motion.y)))*1500.0f
					);
				}
				getCurrentWorld().onMotion(info);
			}
			break;
		}
	}
}


void Game::start()
{
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glShadeModel( GL_SMOOTH );
	glCullFace( GL_BACK );
	glFrontFace( GL_CCW );
	glEnable( GL_CULL_FACE );
	glEnable( GL_BLEND );
	glEnable( GL_DEPTH_TEST );
	glClearColor( 0, 0, 0, 0 );

	font.load(GetGameConfig().system_font);


	for(std::list<std::string>::iterator p=GetGameConfig().dso_list.begin(); p!=GetGameConfig().dso_list.end(); ++p)
		LoadDSO(*p);

#ifdef WIN32
	WIN32_FIND_DATA wfdata;
	HANDLE handle = FindFirstFile("faction/*.dso", &wfdata);
	if(handle!=INVALID_HANDLE_VALUE) {
		do {
			loadDSO(std::string("faction/")+wfdata.cFileName);
		} while(FindNextFile(handle, &wfdata));
		FindClose(handle);
	}
#else
	DIR *dir = opendir("faction");
	if(dir!=0) {
		dirent *dr;
		while((dr = readdir(dir))!=0) {
			if(strlen(dr->d_name)>4 && strcmp(dr->d_name+strlen(dr->d_name)-4, ".dso")==0)
				loadDSO(std::string("faction/")+dr->d_name);
		}
		closedir(dir);
	}
#endif

	const size_t fps = 1000/60;
	size_t start_tick = SDL_GetTicks();
	while(!end_flag) {
		static size_t pre_tick, cur_tick;
		pre_tick = cur_tick;
		cur_tick = SDL_GetTicks();

		if(GetGameConfig().nowait) {
			static size_t last_draw;
			update();
			if(cur_tick-last_draw>=fps) {
				last_draw = cur_tick;
				draw();
			}
		}
		else if(GetGameConfig().frameskip) {
			static size_t c_update, c_draw, skipped;

			bool skip = false;
			size_t norma = (cur_tick-start_tick)/fps;

			if(c_update<norma) {
				update();
				++c_update;
				if(c_update<norma) {
					skip = true;
					++skipped;
				}
			}

		//	if(c_draw<norma && !skip) {
			if(c_draw<norma && (!skip || skipped>30)) {
				draw();
				skipped = 0;
				c_draw = (SDL_GetTicks()-start_tick)/fps;
			}
		}
		else {
			static size_t pre_frame;
			if(cur_tick-pre_frame>=fps) {
				pre_frame = cur_tick;
				update();
				draw();
			}
		}
	}

	for(world_cont::iterator p=worlds.begin(); p!=worlds.end(); ++p)
		p->second->killAll();
	worlds.clear();

	for(dso_cont::iterator p=dsos.begin(); p!=dsos.end(); ++p) {
		typedef void (*OnReleaseFunc)();
		OnReleaseFunc OnRelease = (OnReleaseFunc)p->second->getFunction("OnRelease");
		if(OnRelease!=0)
			OnRelease();
	}
}

void Game::update()
{
	input.update();


	processEvent();
	for(world_cont::iterator p=worlds.begin(); p!=worlds.end(); ++p)
		p->second->update();
	if(!requested_world_name.empty()) {
		setCurrentWorld(requested_world_name);
		requested_world_name.clear();
	}
}

void Game::draw()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	getCurrentWorld().draw();


	static int pre_fps, cur_fps, pre_sec;
	int cur_sec = time(0);

	++cur_fps;
	if(pre_sec!=cur_sec) {
		pre_sec = cur_sec;
		pre_fps = cur_fps;
		cur_fps = 0;
	}

	char text[128];
	static GLubyte white[] = { 255, 255, 255, 255 };
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glColor4ubv( white );
	sprintf(text, "FPS:%d", pre_fps);
	font.print(5, 5, text);

	SDL_GL_SwapBuffers();
}



faction_ptr Game::createFaction(const std::string& name)
{
	faction_ptr p;
	faction_table::iterator it = factiontable.find(name);
	if(it!=factiontable.end())
		p = it->second();
	return p;
}

unit_ptr Game::createUnit(const std::string& name, faction_ptr f)
{
	unit_ptr p;
	unit_table::iterator it = unittable.find(name);
	if(it!=unittable.end())
		p = it->second(f);
	return p;
}

building_ptr Game::createBuilding(const std::string& name, faction_ptr f)
{
	building_ptr p;
	building_table::iterator it = buildingtable.find(name);
	if(it!=buildingtable.end())
		p = it->second(f);
	return p;
}

material_ptr Game::createMaterial(const std::string& name)
{
	material_ptr p;
	material_table::iterator it = materialtable.find(name);
	if(it!=materialtable.end())
		p = it->second();
	return p;
}

group_ptr create_group(IGObjIterator& i);
group_ptr Game::createGroup(IGObjIterator& i)
{
	return create_group(i);
}


void Game::addFactionCreator(const std::string& name, faction_creator c)
{
	assert(factiontable.find(name)==factiontable.end());
	factiontable[name] = c;
	std::cout << "faction " << name << " added." << std::endl;
}

void Game::addUnitCreator(const std::string& name, unit_creator c) {
	assert(unittable.find(name)==unittable.end());
	unittable[name] = c;
}

void Game::addBuildingCreator(const std::string& name, building_creator c) {
	assert(buildingtable.find(name)==buildingtable.end());
	buildingtable[name] = c;
}

void Game::addMaterialCreator(const std::string& name, material_creator c) {
	assert(materialtable.find(name)==materialtable.end());
	materialtable[name] = c;
}






bool Game::loadDSO(const std::string& filename)
{
	if(dsos.find(filename)!=dsos.end())
		return false;

	dso_ptr dso = dso_ptr(new ist::DSO(filename));
	if(!(*dso))
		return false;

	std::cout << filename << " loaded." << std::endl;
	dsos[filename] = dso;

	typedef void (*OnLoadFunc)(IGame&);
	OnLoadFunc OnLoad = (OnLoadFunc)dso->getFunction("OnLoad");
	if(OnLoad==0)
		throw std::runtime_error("function OnLoad() not found");
	OnLoad(*this);

	return true;
}


bool Game::loadXMLDocument(const std::string& filename)
{
	XMLLoader loader;
	return loader.load(filename);
}


void Game::setCurrentWorld(const std::string& name)
{
	if(name=="title") {
		lua_binder_config();
		for(world_cont::iterator p=worlds.begin(); p!=worlds.end();) {
			if(p->first!="title") {
				p->second->killAll();
				worlds.erase(p++);
			}
			else
				++p;
		}
	}

	current_world_name = name;
	world_ptr& wp = worlds[name];
	if(!wp) {
		wp = new World();
		if(name!="title")
			wp->registKeyObserver(SDLK_ESCAPE, keyobserv_ptr(new EscObserver()));
	}
}

void Game::drawQuad(const ist::vector2& v1, const ist::vector2& v2, const ist::vector2& v3, const ist::vector2& v4)
{
	float x_rate = (float)GetFrameConfig().width/640.0f;
	float y_rate = (float)GetFrameConfig().height/480.0f;
	ist::vector2 rel_v1(v1.x*x_rate, (480.0f-v1.y)*y_rate);
	ist::vector2 rel_v2(v2.x*x_rate, (480.0f-v2.y)*y_rate);
	ist::vector2 rel_v3(v3.x*x_rate, (480.0f-v3.y)*y_rate);
	ist::vector2 rel_v4(v4.x*x_rate, (480.0f-v4.y)*y_rate);

	ist::MatrixSaver msaver(GL_MODELVIEW_MATRIX);
	ist::MatrixSaver psaver(GL_PROJECTION_MATRIX);
	ist::ScreenMatrix();

	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(rel_v1.x, rel_v1.y);
	glTexCoord2f(0, 1);
	glVertex2f(rel_v2.x, rel_v2.y);
	glTexCoord2f(1, 1);
	glVertex2f(rel_v3.x, rel_v3.y);
	glTexCoord2f(1, 0);
	glVertex2f(rel_v4.x, rel_v4.y);
	glEnd();
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}

void Game::drawRect(const ist::vector2& min, const ist::vector2& max)
{
	float x_rate = (float)GetFrameConfig().width/640.0f;
	float y_rate = (float)GetFrameConfig().height/480.0f;
	ist::vector2 rel_min(min.x*x_rate, (480.0f-min.y)*y_rate);
	ist::vector2 rel_max(max.x*x_rate, (480.0f-max.y)*y_rate);

	ist::MatrixSaver msaver(GL_MODELVIEW_MATRIX);
	ist::MatrixSaver psaver(GL_PROJECTION_MATRIX);
	ist::ScreenMatrix();

	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex2f(rel_min.x, rel_min.y);
	glTexCoord2f(0, 1);
	glVertex2f(rel_min.x, rel_max.y);
	glTexCoord2f(1, 1);
	glVertex2f(rel_max.x, rel_max.y);
	glTexCoord2f(1, 0);
	glVertex2f(rel_max.x, rel_min.y);
	glEnd();
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}

void Game::drawEdge(const ist::vector2& min, const ist::vector2& max)
{
	ist::vector2 rel_min(min.x, (GetFrameConfig().height-min.y));
	ist::vector2 rel_max(max.x, (GetFrameConfig().height-max.y));

	ist::MatrixSaver msaver(GL_MODELVIEW_MATRIX);
	ist::MatrixSaver psaver(GL_PROJECTION_MATRIX);
	ist::ScreenMatrix();

	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glBegin(GL_LINE_LOOP);
	glTexCoord2f(0, 0);
	glVertex2f(rel_min.x, rel_min.y);
	glTexCoord2f(0, 1);
	glVertex2f(rel_min.x, rel_max.y);
	glTexCoord2f(1, 1);
	glVertex2f(rel_max.x, rel_max.y);
	glTexCoord2f(1, 0);
	glVertex2f(rel_max.x, rel_min.y);
	glEnd();
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
}

void Game::drawFont(int x, int y, const char *message)
{
	float x_rate = (float)GetFrameConfig().width/640.0f;
	float y_rate = (float)GetFrameConfig().height/480.0f;
	font.print(int(x*x_rate), int(y*y_rate), message);
}







