#ifndef DEMO_INCLUDE
#define DEMO_INCLUDE

#include <map>
#include <queue>
#include <iostream>
#include <SDL/SDL.h>
#include "aos/aos.h"

#include "config.h"
#include "input.h"
#include "resource_manager.h"
#include "event_router.h"
#include "world.h"



class Frame : public IGame
{
public:
	Frame();
	virtual ~Frame();
};




class Game : public Frame
{
public:
	typedef boost::intrusive_ptr<ist::DSO> dso_ptr;
	typedef boost::intrusive_ptr<World> world_ptr;
	typedef std::map<std::string, dso_ptr> dso_cont;
	typedef std::map<std::string, world_ptr> world_cont;
	typedef boost::intrusive_ptr<ISoundPlayer> sound_ptr;

	typedef std::map<std::string, faction_creator>  faction_table;
	typedef std::map<std::string, unit_creator>     unit_table;
	typedef std::map<std::string, building_creator> building_table;
	typedef std::map<std::string, material_creator> material_table;


	static Game& instance() {
		static Game g;
		return g;
	}

	void start();
	void update();
	void draw();
	void processEvent();

	bool loadDSO(const std::string& filename);
	bool loadXMLDocument(const std::string& filename);

	faction_ptr  createFaction(const std::string& name);
	unit_ptr     createUnit(const std::string& name, faction_ptr f);
	building_ptr createBuilding(const std::string& name, faction_ptr f);
	material_ptr createMaterial(const std::string& name);
	group_ptr    createGroup(IGObjIterator& i);
	void addFactionCreator(const std::string& name, faction_creator);
	void addUnitCreator(const std::string& name, unit_creator);
	void addBuildingCreator(const std::string& name, building_creator);
	void addMaterialCreator(const std::string& name, material_creator);


	World& getWorld(const std::string& name) { return *worlds[name]; }
	World& getCurrentWorld() { return *worlds[current_world_name]; }
	void setCurrentWorld(const std::string& name);
	void requestSetCurrentWorld(const std::string& name) { requested_world_name=name; }

	bool getAxis(int i)   { return input.getAxis(i); }
	bool getButton(int i) { return input.getButton(i); }
	int getWindowWidth()  { return FrameConfig::instance().width; }
	int getWindowHeight() { return FrameConfig::instance().height; }

	void drawQuad(const ist::vector2& v1, const ist::vector2& v2, const ist::vector2& v3, const ist::vector2& v4);
	void drawRect(const ist::vector2& min, const ist::vector2& max);
	void drawEdge(const ist::vector2& min, const ist::vector2& max);
	void drawFont(int x, int y, const char *message);
	ist::Kanji& getFont()    { return font; }
	float getRand() { return (float)ist::Random::get(); }
	bool isShiftKeyDown() { return is_shift_key_down; }
	bool isAltKeyDown()   { return is_alt_key_down; }
	bool isCtrlKeyDown()  { return is_ctrl_key_down; }

	ResourceManager& getResourceManager() { return  resource_manager; }
	ISoundPlayer& getSoundPlayer()    { return *sound_player; }
	void setSoundPlayer(sound_ptr sp) { sound_player=sp; }

	void enableNoWait(bool f) { GetGameConfig().nowait = f; }
	void exit() { end_flag=true; }

private:
	Game();
	~Game();

	bool is_shift_key_down;
	bool is_alt_key_down;
	bool is_ctrl_key_down;
	bool end_flag;
	ist::Kanji font;

	world_cont worlds;
	std::string current_world_name;
	std::string requested_world_name;

	faction_table  factiontable;
	unit_table     unittable;
	building_table buildingtable;
	material_table materialtable;

	dso_cont dsos;

	Input input;
	ResourceManager resource_manager;
	sound_ptr sound_player;
};




#define GetIGame()       Game::instance()
#define EraseGObj(p)     GetIGame().eraseGObj(p)
#define EraseMenu(p)     GetIGame().eraseMenu(p)

#include "aos/aos.h"


#endif
