#ifndef GROUND
#define GROUND

#include <string>
#include <boost/smart_ptr.hpp>
#include "ist/ist_i3d.h"

inline ist::vector2 ToVector2(const ist::vector4 v4) { return ist::vector2(v4.x, v4.z); }


class IGround
{
public:
	virtual ~IGround() {}

	virtual bool load(const std::string& filename)=0;
	virtual void draw() const=0;
	virtual void draw(const ist::PerspectiveCamera& camera) const=0;
	virtual float getHeight(const ist::vector4& pos) const=0;
	virtual const ist::vector2& getMin() const=0;
	virtual const ist::vector2& getMax() const=0;
};


class ISightMap
{
public:
	virtual ~ISightMap() {}

	virtual void clearSight()=0;
	virtual void addSight(ISightMap& another)=0;
	virtual void addSight(const ist::vector4& pos, float radius)=0;
	virtual bool isSighted(const ist::vector4& pos) const=0;
	virtual bool isExplored(const ist::vector4& pos) const=0;

	virtual size_t getDataLength() const=0;
	virtual const char* getSightData() const=0;
};



class ICollisonMap
{
public:
	virtual ~ICollisonMap() {}
	virtual bool isCollide(const ist::vector4& pos) const=0;
};


template<class T>
class TMapBase : public T
{
public:
	typedef TMapBase BaseClass;
	struct coord_t
	{
		coord_t(int _x=0, int _z=0) : x(_x), z(_z) {}
		coord_t operator-(const coord_t& c) { return coord_t(x-c.x, z-c.z); }
		int x, z;
	};

	virtual void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs) {
		min_edge = new_min;
		max_edge = new_max;
		grid_size = gs;
		x_div = std::max(int((max_edge.x-min_edge.x)/grid_size), 0)+1;
		z_div = std::max(int((max_edge.y-min_edge.y)/grid_size), 0)+1;
	}

	coord_t getCoord(const ist::vector2& pos) const
	{
		ist::vector2 rel_pos = pos-min_edge;
		return coord_t(int(rel_pos.x/grid_size), int(rel_pos.y/grid_size));
	//	return coord_t(int(Round(rel_pos.x/grid_size)), int(Round(rel_pos.y/grid_size)));
	}

	coord_t getClampedCoord(const ist::vector2& pos) const
	{
		coord_t c = getCoord(pos);
		c.x = std::min(getXDiv(), std::max(0, c.x));
		c.z = std::min(getZDiv(), std::max(0, c.z));
		return c;
	}

	coord_t getSafeCoord(const ist::vector2& pos) const
	{
		coord_t c = getCoord(pos);
		c.x = std::min(getXDiv()-1, std::max(0, c.x));
		c.z = std::min(getZDiv()-1, std::max(0, c.z));
		return c;
	}

	const ist::vector2& getMin() const { return min_edge; }
	const ist::vector2& getMax() const {
		aligned_max_edge = getMin()+ist::vector2((x_div-1)*grid_size, (z_div-1)*grid_size);
		return aligned_max_edge;
	}
	int getXDiv() const { return x_div; }
	int getZDiv() const { return z_div; }
	float getGridSize() const { return grid_size; }

	size_t getDataLength() const { return x_div*z_div; }

protected:
	ist::vector2 min_edge;
	ist::vector2 max_edge;
	mutable ist::vector2 aligned_max_edge;
	float grid_size;
	int x_div;
	int z_div;
};



class SightMap : public TMapBase<ISightMap>
{
public:
	enum {
		SIGHTED=1,
		EXPLORED=2,
		OPENED=4,
	};

	void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs) {
		BaseClass::resize(new_min, new_max, gs);
		sight.resize(getXDiv()*getZDiv());
	}

	void clearSight() {
		for(size_t i=0; i<sight.size(); ++i)
			sight[i]&=EXPLORED;
	}

	void fillSight() {
		for(size_t i=0; i<sight.size(); ++i)
			sight[i]|=OPENED;
	}

	void addSight(ISightMap& another) {
		if(another.getDataLength()!=getDataLength())
			return;
		const char *dat = another.getSightData();
		for(size_t i=0; i<sight.size(); ++i)
			sight[i]|=dat[i];
	}

	void addSight(const ist::vector4& pos, float radius) {
		coord_t mn = getClampedCoord(ist::vector2(pos.x-radius, pos.z-radius));
		coord_t mx = getClampedCoord(ist::vector2(pos.x+radius, pos.z+radius));
		ist::vector2 pos2(pos.x, pos.z);
		for(int i=mn.z; i<mx.z; ++i) {
			for(int j=mn.x; j<mx.x; ++j) {
				ist::vector2 vpos = getMin()+ist::vector2(getGridSize()*j, getGridSize()*i);
				if((vpos-pos2).square()<radius*radius) {
					(*this)[i][j]|=SIGHTED;
					(*this)[i][j]|=EXPLORED;
				}
			}
		}
	}


	bool isSighted(const ist::vector4& pos) const {
		return isSighted(getSafeCoord(ist::vector2(pos.x, pos.z)));
	}
	bool isSighted(int x, int z) const {
		return isSighted(coord_t(x, z));
	}
	bool isSighted(const coord_t& c) const {
		char s=(*this)[c.z][c.x];
		return (s&(SIGHTED|OPENED))!=0;
	}

	bool isExplored(const ist::vector4& pos) const {
		return isExplored(getSafeCoord(ist::vector2(pos.x, pos.z)));
	}
	bool isExplored(int x, int z) const {
		return isExplored(coord_t(x, z));
	}
	bool isExplored(const coord_t& c) const {
		char s=(*this)[c.z][c.x];
		return (s&(EXPLORED|OPENED))!=0;
	}

	char* operator[](int i) { return &sight[i*getXDiv()]; }
	const char* operator[](int i) const { return &sight[i*getXDiv()]; }

	const char* getSightData() const { return &sight[0]; }

private:
	std::vector<char> sight;
};




class BGround : public TMapBase<IGround>
{
public:
	struct vertex_t
	{
		vertex_t() : attr(0) {}

		ist::vector3 pos;
		ist::vector3 nml;
		ist::vector4 col;
		ist::vector2 tex;
		int attr;
	};



	typedef std::vector<vertex_t> vertex_cont;
	typedef std::vector<int> index_cont;


	bool load(const std::string& fname);
	void draw() const;
	void draw(const ist::PerspectiveCamera& camera) const;
	float getHeight(const ist::vector4& pos) const;
	float getHeight(const ist::vector2& _target_pos) const;

	BGround();
	void clear();
	void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs);
	void resizePreserved(const ist::vector2& new_min, const ist::vector2& new_max, float gs);
	void setNormal();
	void setNormal(const coord_t& mn, const coord_t& mx);
	void setIndex();
	void initVertex();
	void getHeightRange(float &low, float &high) const;
	ist::vector4 BGround::getRayCross(const ist::vector4 &eye, const ist::vector4 &ray) const;
	vertex_t* operator[](int i) { return &vertex[i*x_div]; }
	const vertex_t* operator[](int i) const { return &vertex[i*x_div]; }



protected:
	vertex_cont vertex;
	index_cont index;
};


class Ground : public BGround
{
public:
	friend class Block;
	class Block
	{
	public:

		Block() : pg(0), updated(false) {}
		void reset(Ground& g, coord_t _mn, coord_t _mx) {
			pg = &g;
			mn = _mn;
			mx = _mx;
			clearVertex();
			for(int i=mn.z; i<mx.z; ++i) {
				for(int j=mn.x; j<mx.x; ++j) {
					addVertex(g[i][j]);
					addVertex(g[i+1][j]);
					addVertex(g[i+1][j+1]);
					addVertex(g[i][j+1]);
				}
			}
			ist::vector3 center = (g[mn.z][mn.x].pos+g[mx.z-1][mx.x-1].pos)/2.0f;
			bs.pos = ist::vector4(center.v);
			bs.radius = (center-g[mn.z][mn.x].pos).norm();
		}

		void clearVertex() {
			tex.clear();
			nml.clear();
			col.clear();
			pos.clear();
		}
		void addVertex(const vertex_t& v) {
			tex.push_back(v.tex);
			nml.push_back(v.nml);
			col.push_back(v.col);
			pos.push_back(v.pos);
		}


		void getNormal(int x, int z, ist::vector3& nml) const {
			const vertex_t& v = (*pg)[z][x];
			if(!pg->sight->isExplored(x, z))
				nml = ist::vector3(0, 0, 0);
			else if(!pg->sight->isSighted(x, z))
				nml = v.nml*0.5f;
			else
				nml = v.nml;
		}
		void setNormal() const {
			int w = mx.x-mn.x;
			int c=0;
			ist::vector3 normal;
			for(int i=mn.z; i<mx.z; ++i) {
				for(int j=mn.x; j<mx.x; ++j) {
					getNormal(j, i, normal);

					nml[c*4] = normal;
					if(c%w!=0)
						nml[(c-1)*4+3] = normal;
					if(c-w>=0) {
						nml[(c-w)*4+1] = normal;
						if(c%w!=0)
							nml[(c-w-1)*4+2] = normal;
					}

					if(j==mx.x-1) {
						getNormal(j+1, i, normal);
						nml[c*4+3] = normal;
						if(c-w>=0)
							nml[(c-w)*4+2] = normal;
					}
					if(i==mx.z-1) {
						getNormal(j, i+1, normal);
						nml[c*4+1] = normal;
						if(c%w!=0)
							nml[(c-1)*4+2] = normal;
					}
					if(j==mx.x-1 && i==mx.z-1) {
						getNormal(j+1, i+1, normal);
						nml[c*4+2] = normal;
					}
					++c;
				}
			}
		}


		void getColor(int x, int z, ist::vector4& color) const {
			const vertex_t& v = (*pg)[z][x];
			if(!pg->sight->isExplored(x, z))
				color = ist::vector4(0, 0, 0, 0);
			else if(!pg->sight->isSighted(x, z))
				color = v.col-ist::vector4(0.5f, 0.5f, 0.5f, 0);
			else
				color = v.col;
		}
		void setColor() const {
			int w = mx.x-mn.x;
			int c=0;
			ist::vector4 color;
			for(int i=mn.z; i<mx.z; ++i) {
				for(int j=mn.x; j<mx.x; ++j) {
					getColor(j, i, color);

					col[c*4] = color;
					if(c%w!=0)
						col[(c-1)*4+3] = color;
					if(c-w>=0) {
						col[(c-w)*4+1] = color;
						if(c%w!=0) {
							col[(c-w-1)*4+2] = color;
						}
					}

					if(j==mx.x-1) {
						getColor(j+1, i, color);
						col[c*4+3] = color;
						if(c-w>=0)
							col[(c-w)*4+2] = color;
					}
					if(i==mx.z-1) {
						getColor(j, i+1, color);
						col[c*4+1] = color;
						if(c%w!=0)
							col[(c-1)*4+2] = color;
					}
					if(j==mx.x-1 && i==mx.z-1) {
						getColor(j+1, i+1, color);
						col[c*4+2] = color;
					}
					++c;
				}
			}
		}

		void draw() const {
 			assert(!!pg && pg->getDataLength()==pg->sight->getDataLength());
			if(!updated) {
				if(pg->isLighting())
					setNormal();
				else
					setColor();
				updated = true;
			}

			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, &tex[0].v);
			if(pg->isLighting()) {
				glEnableClientState(GL_NORMAL_ARRAY);
				glNormalPointer(GL_FLOAT, 0, &nml[0].v);
			}
			else {
				glEnableClientState(GL_COLOR_ARRAY);
				glColorPointer(4, GL_FLOAT, 0, &col[0].v);
			}
			glVertexPointer(3, GL_FLOAT, 0, &pos[0].v);
			glEnableClientState(GL_VERTEX_ARRAY);
			glDrawArrays(GL_QUADS, 0, pos.size());
			glDisableClientState(GL_VERTEX_ARRAY);
			glDisableClientState(GL_COLOR_ARRAY);
			glDisableClientState(GL_NORMAL_ARRAY);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		}
		const ist::Sphere& getBSphere() const { return bs; }

		void setUpdated(bool f) { updated=f; }

	private:
		const Ground* pg;
		mutable bool updated;
		coord_t mn, mx;
		ist::Sphere bs;
		std::vector<ist::vector2> tex;
		mutable std::vector<ist::vector3> nml;
		mutable std::vector<ist::vector4> col;
		std::vector<ist::vector3> pos;
	};



	typedef boost::shared_ptr<SightMap> sight_ptr;
	

	Ground() : divide_size(400), lighting(true), foging(false) {
		light.position.y = 10000.0f;
		light.position.z = -10000.0f;
		sight = createSightMap();
	}

	void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs) {
		BGround::resize(new_min, new_max, gs);
		for(size_t i=0; i<sight_list.size(); ++i)
			sight_list[i]->resize(new_min, new_max, gs);
	}

	void divide() {
		blocks.clear();
		for(int i=0; i<int((getMax().y-getMin().y)/divide_size)+1; ++i) {
			for(int j=0; j<int((getMax().x-getMin().x)/divide_size)+1; ++j) {
				coord_t mn = getSafeCoord(getMin()+ist::vector2(divide_size*j, divide_size*i));
				coord_t mx = getSafeCoord(getMin()+ist::vector2(divide_size*(j+1), divide_size*(i+1)));

				blocks.push_back(Block());
				blocks.back().reset(*this, mn, mx);
			}
		}
	}

	void draw() const {
		glEnable(GL_TEXTURE_2D);
		texture.assign();
		if(lighting) {
			glEnable(GL_LIGHTING);
			light.enable();
		}

		glBegin(GL_QUADS);
		for(size_t i=0; i<index.size(); ++i) {
			glTexCoord2fv(vertex[index[i]].tex.v);
			glNormal3fv(vertex[index[i]].nml.v);
			glVertex3fv(vertex[index[i]].pos.v);
		}
		glEnd();

		light.disable();
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
	}

	void draw(const ist::PerspectiveCamera& camera) const
	{
		glEnable(GL_TEXTURE_2D);
		texture.assign();
		if(lighting) {
			glEnable(GL_LIGHTING);
			light.enable();
		}
		if(foging) {
			fog.enable();
		}
		const ist::ViewFrustum& vf = camera.getViewFrustum();
		for(size_t i=0; i<blocks.size(); ++i) {
			const ist::Sphere& bs = blocks[i].getBSphere();
			if(vf.test(bs.pos, bs.radius))
				blocks[i].draw();
		}
		fog.disable();
		light.disable();
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
	}

	bool load(const std::string& fn) {
		if(BGround::load(fn)) {
			divide();
			return true;
		}
		return false;
	}


	void genBitmap(ist::Bitmap &bmp) {
		for(size_t i=0; i<bmp.getHeight(); ++i) {
			for(size_t j=0; j<bmp.getWidth(); ++j) {
				ist::vector2 s = getMax()-getMin();
				ist::vector2 pos = getMin()+ist::vector2(s.x/bmp.getWidth()*j, s.y/bmp.getHeight()*i);
				ist::vector4 pos4(pos.x, 0, pos.y);
				unsigned char p(0);
				if(!sight->isExplored(pos4))
					p=0;
				else if(!sight->isSighted(pos4))
					p=64;
				else
					p=128;
				bmp[i][j].r=bmp[i][j].g=bmp[i][j].b=p;
				bmp[i][j].a=255;
			}
		}
	}


	ist::Light& getLight() { return light; }
	void loadTexture(const std::string& fn) { texture.load(fn.c_str()); }
	void enableLighting(bool f) { lighting=f; }
	void enableFog(bool f) { foging=f; }
	bool isLighting() const { return lighting; }
	bool isFoging()   const { return foging; }

	sight_ptr createSightMap() {
		sight_ptr t_sight(new SightMap());
		t_sight->resize(getMin(), getMax(), getGridSize());
		sight_list.push_back(t_sight);
		return t_sight;
	}

	sight_ptr getSightMap() {
		return sight;
	}

	void assignSightMap(sight_ptr sp) {
		if(!sp)
			return;
		sight=sp;
		for(size_t i=0; i<blocks.size(); ++i)
			blocks[i].setUpdated(false);
	}

private:
	sight_ptr sight;
	std::vector<sight_ptr> sight_list;
	float divide_size;
	std::vector<Block> blocks;
	bool lighting;
	bool foging;
	mutable ist::Light light;
	ist::Fog fog;
	ist::Texture texture;
};






#endif
