#ifndef WORLD_INCLUDE
#define WORLD_INCLUDE

#include <set>
#include <map>
#include <string>
#include "aos/aos.h"
#include "aos/inlines.h"

#include "ist/grid.h"
#include "event_router.h"

#include "ground.h"
#include "path.h"


class BlockSet : public IGObjIterator
{
public:
	typedef ist::grid_structure2<gobj_ptr> character_cont;
	typedef character_cont::block_elem block_elem;
	typedef character_cont::block_iter block_iter;

	BlockSet() : bs(0) {}
	void set(block_iter& bsp) { bs=&bsp; reset(); }
	void reset() { bs->reset(); }
	bool hasNext() { return bs->hasNext(); }
	gobj_ptr iterate() { return bs->iterate(); }

	block_iter *bs;
};








class World : public IWorld
{
public:
	World();
	~World();

	void setMode(int mode);
	int getMode();
	bool loadReplay(const std::string& filename);
	void setServerAddress(const std::string& address);
	void setServerPort(unsigned short address);
	const std::string& getServerDatabase();

	void registKeyObserver(int key, keyobserv_ptr observer);
	void deregistKeyObserver(int key, keyobserv_ptr observer);

	gobj_id createID();
	void setIDSeed(size_t ids) { id_seed=ids; }
	faction_ptr  createFaction(const std::string& name);
	unit_ptr     createUnit(const std::string& name, faction_ptr f);
	building_ptr createBuilding(const std::string& name, faction_ptr f);
	material_ptr createMaterial(const std::string& name);
	group_ptr    createGroup(IGObjIterator& i);

	gobj_ptr getGObj(gobj_id id);

	bool addGObj(gobj_ptr p);
	void eraseGObj(gobj_ptr p);

	void insertReciever(gobj_ptr p);
	void eraseReciever(gobj_ptr p);

	void update();
	void draw();

	void killAll();

	void onButton(Event::buttoninfo_ptr info);
	void onMotion(Event::motioninfo_ptr info);
	void onKey(Event::keyinfo_ptr info);
	void sendUpdateEvent();
	void stopSendingUpdateEvent() { is_update_stoped=true; }
	void startSendingUpdateEvent(){ is_update_stoped=false; }

	void loadGround(const std::string& filename);
	void loadGroundTexture(const std::string& filename);
	float getHeight(const ist::vector4& pos);
	const ist::vector2& getMapMin();
	const ist::vector2& getMapMax();
	ist::Light& getGroundLight();
	void enableGroundLighting(bool f);
	void enableGroundFog(bool f);
	void getMapTexture(ist::Bitmap& bmp);
	bool isSighted(faction_ptr f, const ist::vector4& pos);
	bool isExplored(faction_ptr f, const ist::vector4& pos);
	void enableSight(bool s);
	IPathFinder& getPathFinder() { return pfinder; }


	IGObjIterator& getBlocksByRadius(const ist::vector4& pos, float radius);
	IGObjIterator& getBlocksByBox(const ist::vector4& min, const ist::vector4& max);
	IGObjIterator& getGObjsInSight();
	IGObjIterator& getFactions();
	IGObjIterator& getMenus();

	void insertToCharacterGrid(const ist::vector4& pos, gobj_ptr p);
	void eraseFromCharacterGrid(const ist::vector4& pos, gobj_ptr p);
	void moveCharacterGrid(const ist::vector4& before, const ist::vector4& after, gobj_ptr p);
	void addCircleCollision(selectable_ptr p, const ist::vector4& pos, float radius);
	void removeCircleCollision(selectable_ptr p, const ist::vector4& pos, float radius);

	void sendEvent(const Event& e) { event_router->send(e); }

	void setPlayerFaction(faction_ptr p);
	void setNatureFaction(faction_ptr p);
	faction_ptr getPlayerFaction();
	faction_ptr getNatureFaction();


	ist::PerspectiveCamera& getCamera() { return camera; }
	const ist::vector4& getCameraPos() { return camera.target; }
	ist::matrix44& getAimCameraMatrix() { return aim_camera; }
	const ist::vector4& getScreenRay() { return screen_ray; }
	const ist::vector4& getMousePointerPos3D() { return mouse_pointer_pos3; }
	const ist::vector2& getMousePointerPos2D() { return mouse_pointer_pos2; }
	void setCameraPos(const ist::vector4&);

	ist::vector4 getScreenRay(const ist::vector2&);
	ist::vector4 _getMousePointerPos3D();
	void moveCamera();

	frame_t getCurrentFrame() { return frame; }

	void setFrameLimit(frame_t f) { frame_limit=f; }

//	typedef grid_structure<gobj_ptr> character_cont;
	typedef ist::grid_structure2<gobj_ptr> character_cont;
	typedef std::vector<gobj_ptr> gobj_vector;
	typedef std::vector<gobj_ptr> draw_cont;
	typedef std::set<gobj_ptr, less_id<gobj_ptr> > gobj_set;
	typedef std::set<menu_ptr, less_id<menu_ptr> > menu_set;
	typedef std::set<faction_ptr, less_id<faction_ptr> > faction_set;
	typedef std::map<gobj_id, gobj_ptr> gobj_cont;
	typedef std::list<gobj_ptr> gobj_list;
	typedef std::map<faction_ptr, Ground::sight_ptr> sight_cont;

	typedef std::list<keyobserv_ptr> keyob_list;
	typedef std::map<int, keyob_list> keymap_cont;

private:
	gobj_id id_seed;

	erouter_ptr event_router;
	keymap_cont keymap;

	BlockSet blocks;

	frame_t frame;
	frame_t frame_limit;
	character_cont character_grid;
	Ground ground;
	PathFinder pfinder;
	draw_cont drawlist;

	gobj_cont gobjs;
	gobj_list update_list;
	menu_set menus;
	faction_set factions;
	gobj_vector gobjs_insight;
	gobj_vector gobjs_insight_selectable;

	gobj_set rcv_update;
	gobj_set rcv_keyboard;

	faction_ptr player_faction;
	faction_ptr nature_faction;
	sight_cont sights;
	bool enable_sight;

	ist::Light ground_light;
	ist::PerspectiveCamera camera;
	ist::vector4 camera_dist;

	ist::matrix44 aim_camera;
	ist::vector4 screen_ray;
	ist::vector4 mouse_pointer_pos3;
	ist::vector2 mouse_pointer_pos2;

	bool is_update_stoped;
};


#endif
