#ifndef EQUIPMENT_INCLUDE
#define EQUIPMENT_INCLUDE

#include "game_object.h"


class IEquipment : public ist::Object
{
public:
	virtual ~IEquipment() {}

	virtual bool isDead()=0;
	virtual void update()=0;
	virtual void draw()=0;
	virtual void equip(breakable_ptr root)=0;
	virtual void unequip()=0;
	virtual void setPos(const ist::vector4& pos)=0;
	virtual const std::string& getName()=0;
	virtual const std::string& getDescription()=0;
	virtual const std::string& getClassName()=0;

	virtual bool isWeapon() { return false; }
	virtual bool isArmor()  { return false; }
	virtual bool isMisc()   { return false; }
};


class IWeapon : public IEquipment
{
public:
	virtual bool ready()=0;
	virtual void fire(selectable_ptr target)=0;
	virtual float getFirePower()=0;
	virtual float getRange()=0;
	virtual int getCoolDown()=0;

	bool isWeapon() { return true; }
};

class IArmor : public IEquipment
{
public:
	virtual void damage(bullet_ptr bullet)=0;
	virtual float getAntiBullet()=0;
	virtual float getAntiLaser()=0;
	virtual float getAntiBomb()=0;

	bool isArmor() { return true; }
};



typedef boost::intrusive_ptr<IEquipment> equipment_ptr;
typedef boost::intrusive_ptr<IWeapon> weapon_ptr;
typedef boost::intrusive_ptr<IArmor> armor_ptr;

inline weapon_ptr ToWeapon(equipment_ptr p) {
	return (!!p && p->isWeapon()) ? boost::static_pointer_cast<IWeapon>(p) : weapon_ptr();
}
inline armor_ptr ToArmor(equipment_ptr p) {
	return (!!p && p->isArmor()) ? boost::static_pointer_cast<IArmor>(p) : armor_ptr();
}



template <class T>
class TEquipment : public T
{
public:
	TEquipment() :
		life(1.0f)
	{}
	~TEquipment() {
		if(!!root)
			unequip();
	}
	void equip(breakable_ptr r) { root=r; }
	void unequip() { root=breakable_ptr(); }
	void setPos(const ist::vector4& p) { pos=p; }

protected:
	breakable_ptr getRoot() { return root; }
	const ist::vector4& getPos() { return pos; }
	float getLife() { return life; }
	void setLife(float f) { life=f; }

private:
	ist::vector4 pos;
	breakable_ptr root;
	float life;
};


class Weapon : public LinearInherit2(TEquipment, IWeapon)
{
public:
	Weapon() :
		fire_power(1.0f),
		fire_range(0.0f),
		cool_down(0)
	{}

	virtual bullet_ptr createBullet()=0;

	bool ready() {
		if(getCoolDown()>0 || !!getRoot())
			return false;
		return true;
	}

	void fire(selectable_ptr target) {
		breakable_ptr r = getRoot();
		bullet_ptr b = createBullet();
		ist::matrix44 mat;
		r->calcMatrix(mat);
		b->setPos(mat*getPos());
		AddGObj(b);
	}

	void update() {
		if(cool_down>0)
			--cool_down;
	}

	ist::vector4 getFirePos() {
		if(breakable_ptr r = getRoot()) {
			ist::matrix44 mat;
			r->calcMatrix(mat);
			return mat*getPos();
		}
		return ist::vector4();
	}

	float getFirePower() { return fire_power; }
	float getFireRange() { return fire_range; }
	int getCoolDown()    { return cool_down; }

protected:
	void setFirePower(float f) { fire_power=f; }
	void setFireRange(float f) { fire_range=f; }
	void serCoolDown(int i)    { cool_down=i; }

private:
	float fire_power;
	float fire_range;
	int cool_down;
};


#endif
